"""vdb.py: Module providing functions for reading from VDB datasets."""

# Nexus is a registered trademark of the Halliburton Company

import logging

log = logging.getLogger(__name__)

import glob
import os
import zipfile as zf
import numpy as np
from struct import unpack

import resqpy.olio.grid_functions as gf
import resqpy.olio.xml_et as rqet

null_uint32 = 4294967295  # -1 if interpreted as int32

key_dict = {  # vdb key character mapping to: (numpy_dtype, size_in_bytes, unpack_format_ch)
    'R': ('float32', 4, 'f'),
    'D': ('float64', 8, 'd'),
    #   'D': ('int64',   8, 'i'),
    'I': ('int32', 4, 'i'),
    'C': (None, 1, 'c'),  # could map to numpy 'byte' but seems to be used for strings
    'P': ('uint32', 4, 'I'),
    'K': (None, 8, 'c'),  # don't store in numpy format; 8 character strings
    'X': (None, 0, 'c')
}  # used for invalid code character (non-ascii)

init_not_packed = ['DAD', 'KID', 'UID', 'UNPACK']


def coerce(a, dtype):
    """Returns a version of numpy array a with elements coerced to dtype.

    :meta private:
    """
    if dtype is None or a.dtype == dtype:
        return a
    b = np.empty(a.shape, dtype = dtype)
    b[:] = a
    return b


def ensemble_vdb_list(run_dir, sort_list = True):
    """Returns a sorted list of vdb paths found in the directory tree under run_dir."""

    ensemble_list = []

    def recursive_vdb_list(dir):
        nonlocal ensemble_list
        for entry in os.scandir(dir):
            if not entry.is_dir():
                continue
            if entry.name.endswith('.vdb') or entry.name.endswith('.vdb.zip'):
                ensemble_list.append(entry.path)
                continue
            elif entry.name.endswith('.rst'):
                continue  # optimisation
            recursive_vdb_list(entry.path)

    def cmp_to_key(mycmp):
        """Convert a cmp= function into a key= function."""

        class K:

            def __init__(self, obj, *args):
                self.obj = obj

            def __lt__(self, other):
                return mycmp(self.obj, other.obj) < 0

            def __gt__(self, other):
                return mycmp(self.obj, other.obj) > 0

            def __eq__(self, other):
                return mycmp(self.obj, other.obj) == 0

            def __le__(self, other):
                return mycmp(self.obj, other.obj) <= 0

            def __ge__(self, other):
                return mycmp(self.obj, other.obj) >= 0

            def __ne__(self, other):
                return mycmp(self.obj, other.obj) != 0

        return K

    def comparison(a, b):
        """Comparison function for a pair of case names."""
        pa = 0
        pb = 0
        while True:
            sa = pa
            sb = pb
            if pa >= len(a) and pb >= len(b):
                return 0
            while pa < len(a) and not a[pa].isdigit():
                pa += 1
            while pb < len(b) and not b[pb].isdigit():
                pb += 1
            if a[sa:pa].lower() < b[sb:pb].lower():
                return -1
            if a[sa:pa].lower() > b[sb:pb].lower():
                return 1
            if pa >= len(a) and pb >= len(b):
                return 0
            if pa >= len(a):
                return -1
            if pb >= len(b):
                return 1
            sa = pa
            sb = pb
            while pa < len(a) and a[pa].isdigit():
                pa += 1
            while pb < len(b) and b[pb].isdigit():
                pb += 1
            ia = int(a[sa:pa])
            ib = int(b[sb:pb])
            if ia < ib:
                return -1
            if ia > ib:
                return 1
        a_low = a.lower()
        b_low = b.lower()
        if a_low < b_low:
            return -1
        if a_low > b_low:
            return 1
        return 0

    recursive_vdb_list(run_dir)
    if sort_list:
        sorted_list = sorted(ensemble_list, key = cmp_to_key(comparison))
        ensemble_list = sorted_list
    return ensemble_list


class Header():
    """Internal class for handling a Header record in a vdb file."""

    def __init__(self, fp, place):
        """Creates a new Header record object."""

        fp.seek(place)
        block = fp.read(22)
        #      log.debug(f'Header block at place {place} returned {len(block)} bytes')
        self.previous, self.next, c, self.bytes_per_item, self.number_of_items, self.first_fragment, self.max_items = \
            unpack('=IIcBIII', block)
        try:
            self.item_type = c.decode()
        except Exception:
            self.item_type = 'X'  # non-ascii character!
        self.data_place = place + 22


#      log.debug('   header previous: ' + str(self.previous))
#      log.debug('   header next: ' + str(self.next))
#      log.debug('   header item type: ' + str(self.item_type))
#      log.debug('   header bytes per item: ' + str(self.bytes_per_item))
#      log.debug('   header number of items: ' + str(self.number_of_items))
#      log.debug('   header first fragment: ' + str(self.first_fragment))
#      log.debug('   header max items: ' + str(self.max_items))
#      log.debug('   header data place: ' + str(self.data_place))


class FragmentHeader():
    """Internal class for handling a Fragment Header record in a vdb file."""

    def __init__(self, fp, place):
        """Creates a new Fragment Header record object."""

        #      log.debug(f'FragmentHeader init at place {place}')
        fp.seek(place)
        block = fp.read(8)
        self.next, self.number_of_items = unpack('=II', block)
        self.data_place = place + 8


#      log.debug(f'   fragment header number of items {self.number_of_items}; data place {self.data_place}; next {self.next}')


class RawData():
    """Internal class for handling Raw Data records in a vdb file."""

    def __init__(self, fp, place, item_type, count, max_count):
        """Creates a new Raw Data record object."""

        if max_count is not None and count > max_count:
            count = max_count
        self.a = None
        self.c = None
        dtype, byte_size, form_ch = key_dict[item_type]
        #      log.debug('raw data call: place {}; item type {}; count {}; dtype {}, byte size {}, form ch {}'.format(
        #                place, item_type, count, dtype, byte_size, form_ch))
        fp.seek(place)
        if dtype is None:
            if form_ch == 'c':
                chars = fp.read(count * byte_size)
                if item_type == 'C':
                    self.c = chars.decode()
                else:
                    self.c = []
                    for i in range(count):
                        self.c.append(chars[i * byte_size:(i + 1) * byte_size].decode().strip().upper())
            else:  # shouldn't come into play
                block = fp.read(count * byte_size)
                form = '=' + str(count) + form_ch
                self.c = unpack(form, block)
        elif dtype == 'float64':  # try tentative 32bit word swap
            b = fp.read(count * 8)
            #         c = b''
            #         for d in range(count):
            #            c += b[4*d+4:4*d+8] + b[4*d:4*d+4]
            nda = np.ndarray((count, 2), dtype = 'int32', buffer = b)
            self.a = np.empty((count, 2), dtype = int)
            self.a[:, :] = nda[:, :]
        else:
            b = fp.read(count * byte_size)
            self.a = np.frombuffer(buffer = b, dtype = dtype, count = count)


class Data():
    """Internal class for handling Data records in a vdb file."""

    def __init__(self, fp, header):
        """Creates a new Data object."""

        if header is None:
            self.a = None
            self.c = None
        else:
            #         log.debug(f'Data init at place {header.data_place}; type {header.item_type}; number of items {header.number_of_items}')
            raw = RawData(fp, header.data_place, header.item_type, header.number_of_items, header.max_items)
            if raw is not None and raw.a is not None and raw.a.size == 1 and raw.a.dtype == 'int32' and header.next != null_uint32:
                #            log.debug('   skipping integer value of ' + str(raw.a[0]))
                next_head = Header(fp, header.next)
                raw = RawData(fp, next_head.data_place, next_head.item_type, next_head.number_of_items,
                              next_head.max_items)
                header = next_head  # making this up as I go along
            self.a = raw.a
            self.c = raw.c
            if header.first_fragment != null_uint32:
                #            log.debug(f'   chaining from {header.first_fragment}')
                chain = FragmentChain(fp, header.first_fragment, header)
                if self.a is not None:
                    self.a = np.append(self.a, chain.a)
                if self.c is not None:
                    self.c += chain.c

    #   if self.c is not None:
    #      log.debug(f'   data c {self.c}')
    #   elif self.a is not None:
    #      log.debug(f'   data a {self.a}')


class Fragment():
    """Internal class for handling an individual Fragment record in a vdb file."""

    def __init__(self, fp, place, header):
        """Creates a new Fragment record object."""

        #      log.debug(f'Fragment init at place {place}')
        self.head = FragmentHeader(fp, place)
        if self.head.number_of_items == 0:
            #         log.debug('   zero items in fragment')
            self.c = None
            self.a = None
        else:
            raw = RawData(fp, self.head.data_place, header.item_type, self.head.number_of_items, header.max_items)
            self.a = raw.a
            self.c = raw.c


#      if self.c is not None:
#         log.debug(f'   fragment c [{self.c}]')


class FragmentChain():
    """Internal class for handling a chain of Fragment records in a vdb file."""

    def __init__(self, fp, place, header):  # returns either 1D numpy array (numeric data) or list of 8
        """Creates a new Fragment Chain object."""

        self.a = None
        self.c = None
        #      log.debug(f'FragmentChain init at place {place}')
        while place != null_uint32:
            fragment = Fragment(fp, place, header)
            if fragment.head.number_of_items > 0:
                if fragment.a is not None:
                    if self.a is None:
                        self.a = fragment.a
                    else:
                        self.a = np.concatenate(self.a, fragment.a)
                if fragment.c:
                    if self.c is None:
                        self.c = []
                    if isinstance(fragment.c, str):
                        self.c.append(fragment.c)
                    elif isinstance(fragment.c, list):
                        for item in fragment.c:
                            if item.isascii():
                                self.c.append(item)
            place = fragment.head.next
        assert self.c is None or self.a is None, 'mixture of character and numeric data in fragment chain'

    # todo: check number of elements matches header info?


class KP():
    """Internal class for a (Key, Pointer) record in a vdb file."""

    def __init__(self, fp, place = 4):
        """Creates a new (Key, Pointer) record object."""

        #      log.debug(f'KP init at place {place}')
        self.k_head = Header(fp, place)
        assert self.k_head.item_type == 'K', 'did not find expected Key header'
        assert self.k_head.bytes_per_item == 8, 'bytes per item not 8 in Key header'
        assert self.k_head.number_of_items > 0, 'zero items in Key header'
        place = self.k_head.next
        assert place != null_uint32, 'no next header in Key header'
        self.p_head = Header(fp, place)
        assert self.p_head.item_type == 'P', 'did not find expected Pointer header'
        assert self.p_head.bytes_per_item == 4, 'bytes per item not 4 in Pointer header'
        assert self.p_head.number_of_items == self.k_head.number_of_items, \
            'number of items in Pointer header does not match number in Key header'
        self.keywords = Data(fp, self.k_head).c
        assert self.keywords is not None and isinstance(self.keywords,
                                                        list), 'list of keywords not extracted from Key record'
        self.pointers = Data(fp, self.p_head).a
        assert self.pointers is not None and self.pointers.dtype == 'uint32'
        self.fp = fp

    def header_place_for_key(self, key, search = False):
        """Returns file position pointer for a given key."""

        key = key.strip().upper()
        if key in self.keywords:
            return self.pointers[self.keywords.index(key)]
        if not search:
            return None
        for head_key in self.keywords:
            sub_head_place = self.pointers[self.keywords.index(head_key)]
            try:
                sub_kp = KP(self.fp, sub_head_place)
                sub_place = sub_kp.header_place_for_key(key, search = True)
                if sub_place is not None:
                    return sub_place
            except AssertionError:
                pass  # probably not a K type subsidiary
            except Exception:
                raise
        return None

    def head_for_key(self, key, search = False):
        """Returns a Header record object for the given key."""

        head_place = self.header_place_for_key(key, search = search)
        if head_place is None:
            return None
        head = Header(self.fp, head_place)
        if head.item_type == 'X':
            return None  # invalid (non-ascii) item type character
        return head

    def data_for_key(self, key, search = False):
        """Returns a Data object for the given key."""

        head = self.head_for_key(key, search = search)
        if head is None:
            return None
        return Data(self.fp, head)

    def key_list(self, filter = False):
        """Returns a list of keys."""

        # return Data(self.fp, self.k_head).c
        if not filter:
            return self.keywords
        filtered_list = []
        for key in self.keywords:
            # log.debug(f'key_list raw entry: {key}')
            if not bad_keyword(key):
                filtered_list.append(key)
        return filtered_list

    def sub_key_list(self, keyword, filter = False):
        """Returns a list of keys subordinate to keyword."""

        assert keyword in self.key_list(), 'keyword not present: ' + keyword
        sub_head_place = self.pointers[self.keywords.index(keyword)]
        sub_kp = KP(self.fp, sub_head_place)
        return sub_kp.key_list(filter = filter)


class VDB():
    """Class for handling a vdb, particularly to support import of grid and properties."""

    def __init__(self, path):
        """Initialises a VDB object and associates it with the given vdb directory path."""

        self.path = None  # internal zip path of vdb in the case of a zip file
        self.zip_file = None
        self.zipped = zf.is_zipfile(path)
        if self.zipped:
            self.zip_file = path
            with zf.ZipFile(path) as zfp:
                zip_list = zfp.namelist()
                warn = False
                for name in zip_list:
                    if name.endswith('.vdb/'):
                        if self.path is not None:
                            warn = True
                        self.path = name[:-1]
                    elif self.path is None and name == 'main.xml':
                        self.path = ''
            if warn:
                log.warning('more than one vdb in zip file: using last')
            if self.path is None:
                raise Exception('no vdb found in zip file (at top level)')
        else:
            self.path = path
            if not os.path.isdir(path):
                log.warning('vdb path is neither a directory nor a zip file: ' + str(path))
        self.case_list = None
        self.use_case = None
        self.grid_extents_kji = {}  # maps case name, grid_name to grid extent
        self.grid_packings = {}  # maps case name, grid_name to grid unpack array
        self.grid_lists = {}  # maps case name to grid name list
        self.cases()  # sets use_case to first case in list for vdb

    def cases(self):
        """Returns a list of simulation case strings as found in the main xml file."""

        if self.case_list is not None:
            return self.case_list
        try:
            xml_file = os.path.join(self.path, 'main.xml')
            if self.zipped:
                with zf.ZipFile(self.zip_file) as zfp:
                    with zfp.open(xml_file) as fp:
                        tree = rqet.parse(fp)
                        root = tree.getroot()
            else:
                assert os.path.exists(xml_file), 'could not find vdb main xml file: ' + xml_file
                with open(xml_file, 'r') as fp:
                    tree = rqet.parse(fp)
                    root = tree.getroot()
            caselist = rqet.list_of_tag(rqet.find_tag(root, 'CASELIST'), 'CASE')
            self.case_list = []
            for case in caselist:
                self.case_list.append(str(case.attrib['Name']).strip())
            if self.use_case is None and len(self.case_list):
                self.use_case = self.case_list[0]
        except Exception:
            log.exception('failed to extract case list')
        return self.case_list

    def set_use_case(self, case):
        """Sets the simulation case to use in other functions."""

        try:
            assert case in self.cases(), 'failed to find case ' + case + ' in list of vdb cases'
            self.use_case = case
        except Exception:
            log.exception('failed to set use case')

    def print_header_tree(self, relative_path):
        """Low level: prints out the raw header tree found in the vdb file (for debugging)."""

        visited = []

        def print_tree(fp, place = 4, level = 0, dots = False, file_size = None):
            """Recursive function for printing vdb tree."""
            try:
                if file_size is None:
                    fp.seek(0, 2)
                    file_size = int(fp.tell())
                assert 0 <= int(place) < file_size, f'place {place} not within bounds of file of size {file_size}'
                assert place not in visited, 'circular pointers in vdb file header structure'
                visited.append(place)
                head = Header(fp, place = place)
                if head is None:
                    raise ValueError()
                if dots:
                    indent = ((level - 1) * 3) * ' ' + ' ..'
                else:
                    indent = (level * 3) * ' '
                print(indent,
                      head.item_type,
                      ':',
                      head.number_of_items,
                      'items at',
                      head.bytes_per_item,
                      'bytes per item',
                      end = '')
                if head.item_type == 'C':
                    s = RawData(fp, head.data_place, 'C', head.number_of_items, head.max_items)
                    if s is None or s.c is None:
                        print(': ?', end = '')
                    else:
                        print(': "' + s.c + '"', end = '')
                if head.first_fragment != null_uint32:
                    print(' with fragment(s)')
                else:
                    print('')
                if head.item_type == 'P':
                    pointers = RawData(fp, head.data_place, 'P', head.number_of_items, head.max_items)
                    assert (pointers is not None and pointers.a is not None)
                    for p in pointers.a:
                        if p in visited:
                            log.warning('breaking at circular pointer to place ' + str(p))
                            break
                        if int(p) == 0:
                            log.warning('pointer to place zero encountered')
                        elif int(p) > file_size:
                            log.warning('skipping pointer to place beyond end of file ' + str(p))
                        else:
                            print_tree(fp, place = p, level = level + 1, dots = True, file_size = file_size)
                if head.next != null_uint32:
                    if not 0 <= int(head.next) < file_size:
                        log.warning(f'skipping next {head.next} outwith bounds of file of size {file_size}')
                    else:
                        print_tree(fp, place = head.next, level = level, file_size = file_size)
            except Exception:
                print('?')
                log.exception('failed in print tree for headers')
                raise

        try:
            path = os.path.join(self.path, self.use_case, relative_path)
            assert os.path.exists(path), 'failed to find vdb file ' + path
            print('Header tree for: ' + path)
            with open(path, 'rb') as fp:
                print_tree(fp)
        except Exception:
            log.exception('failed to print header tree for relative file ' + relative_path)

    def print_key_tree(self, relative_path):
        """Low level: prints out the keyword tree found in the vdb file."""

        visited = []

        def print_tree(fp, place = 4, level = 0):
            try:
                if int(place) == 0:
                    log.warning('returning due to place zero in vdb file key structure')
                    return
                if place in visited:
                    log.warning(f'returning at circular pointer at place {place} in vdb file key structure')
                    return
                visited.append(place)
                kp = KP(fp, place = place)
            except AssertionError:
                return
            except Exception as e:
                log.error(str(e) + ' at place ' + str(place))
                raise
            key_list = kp.key_list()
            for key in key_list:
                print((level * 3) * ' ', key)
                print_tree(fp, kp.header_place_for_key(key, search = False), level + 1)

        try:
            path = os.path.join(self.path, self.use_case, relative_path)
            if self.zipped:
                with zf.ZipFile(self.zip_file) as zfp:
                    with zfp.open(path) as fp:
                        print_tree(fp)
            else:
                assert os.path.exists(path), 'failed to find vdb file ' + str(path)
                with open(path, 'rb') as fp:
                    print_tree(fp)
        except Exception:
            log.exception('failed to print keyword tree for relative file ' + relative_path)

    def data_for_keyword(self, relative_path, keyword, search = True):
        """Reads data associated with a keyword from a vdb binary file; returns a numpy array (or string)."""

        try:
            path = os.path.join(self.path, self.use_case, relative_path)
            if self.zipped:
                with zf.ZipFile(self.zip_file) as zfp:
                    with zfp.open(path) as fp:
                        assert fp.read(4) == b'NT32', 'first 4 characters not NT32 in file ' + path
                        kp = KP(fp)
                        return kp.data_for_key(keyword, search = search)
            else:
                assert os.path.exists(path), 'failed to find vdb file ' + str(path)
                with open(path, 'rb') as fp:
                    assert fp.read(4) == b'NT32', 'first 4 characters not NT32 in file ' + path
                    kp = KP(fp)
                    return kp.data_for_key(keyword, search = search)
        except Exception:
            log.exception('failed to read keyword data from vdb binary file: ' + path)
        return None

    def data_for_keyword_chain(self, relative_path, keyword_chain):
        """Follows a list of keywords down through hierarchy and returns the data as a numpy array (or string)."""

        def dfkc_fp(fp, path, keyword_chain):
            assert fp.read(4) == b'NT32', 'first 4 characters not NT32 in file ' + path
            chain_list = list(keyword_chain)
            assert len(chain_list) > 0, 'empty chained keyword list'
            head_place = 4
            while len(chain_list) > 1:
                kp = KP(fp, place = head_place)
                head_place = kp.header_place_for_key(chain_list[0], search = False)
                assert head_place is not None, 'failed to find chained keyword: ' + chain_list[0]
                chain_list.pop(0)
            kp = KP(fp, place = head_place)
            return kp.data_for_key(chain_list[0], search = False)

        path = os.path.join(self.path, self.use_case, relative_path)
        try:
            if isinstance(keyword_chain, str):
                return self.data_for_keyword(relative_path, keyword_chain)
            if self.zipped:
                with zf.ZipFile(self.zip_file) as zfp:
                    with zfp.open(path) as fp:
                        return dfkc_fp(fp, path, keyword_chain)
            else:
                assert os.path.exists(path), 'failed to find vdb file ' + str(path)
                with open(path, 'rb') as fp:
                    return dfkc_fp(fp, path, keyword_chain)
        except Exception:
            log.exception('failed to read chained keyword data from vdb binary file: ' + path)
        return None

    def set_extent_kji(self, extent_kji, use_case = None, grid_name = 'ROOT'):
        """Sets extent for one use case (defaults to current use case) as alternative to processing corp data."""

        assert len(extent_kji) == 3, 'triple integer required for extent_kji'
        if use_case is None:
            use_case = self.use_case
        assert use_case is not None, 'no use case for extent setting'
        self.grid_extents_kji[use_case, grid_name.upper()] = extent_kji

    def fetch_corp_patch(self, relative_path):
        """Loads one patch of grid corp data from one file in the vdb; returns (number of cells, 1D array)."""

        a = self.data_for_keyword(relative_path, 'CORP').a
        assert a is not None, 'failed to extract corp data from vdb relative path ' + relative_path
        element_count = a.size
        assert element_count > 0, 'corp data is empty (no numbers)'
        cells, remainder = divmod(element_count, 24)
        assert remainder == 0, 'number of numbers in corp data is not a multiple of 24'
        if a.dtype == 'float32':
            d = np.empty((element_count,), dtype = 'float')  # double precision
            d[:] = a
            del a
            a = d
        return cells, a

    def list_of_grids(self):
        """Returns a list of grid names for which corp data exists in the vdb for the current use case."""

        if self.use_case is None:
            self.cases()
        if self.use_case in self.grid_lists:
            return self.grid_lists[self.use_case]
        glob_path = os.path.join(self.path, self.use_case, 'INIT', '*_corp.bin')
        if self.zipped:
            corp_list = self.zip_glob(glob_path)
        else:
            corp_list = glob.glob(glob_path)
        grid_name_list = []
        for corp_path in corp_list:
            _, corp_file_name = os.path.split(corp_path)
            grid_name = corp_file_name[:-9].upper()
            if grid_name == 'ROOT':
                grid_name_list.insert(0, 'ROOT')
            else:
                grid_name_list.append(grid_name)
        self.grid_lists[self.use_case] = grid_name_list
        return grid_name_list

    def root_corp(self):
        """Loads root grid corp data from vdb; returns pagoda style resequenced 7D numpy array of doubles."""

        return self.grid_corp('ROOT')

    def grid_corp(self, grid_name):
        """Loads corp data for named grid from vdb; returns pagoda style resequenced 7D numpy array of doubles."""
        try:
            if self.use_case is None:
                self.cases()
            grid_name = grid_name.upper()
            assert self.use_case is not None, 'no case found in vdb'
            relative_path = os.path.join('INIT', str(grid_name) + '_corp.bin')
            cells, a = self.fetch_corp_patch(relative_path)
            # check for extra corp patch files (used for big grids)
            glob_path = os.path.join(self.path, self.use_case, 'INIT', str(grid_name) + '_corp')
            glob_path += '_*.bin'
            if self.zipped:
                corp_patch_list = self.zip_glob(glob_path)
            else:
                corp_patch_list = glob.glob(glob_path)
            if len(corp_patch_list) > 0:
                sort_list = []
                for patch_name in corp_patch_list:
                    layer_number = int(patch_name[patch_name.rfind('_') + 1:-4])
                    sort_list.append((layer_number, patch_name))
                sort_list.sort()
                for _, patch_name in sort_list:
                    patch_relative_path = patch_name[patch_name.rfind('INIT'):]
                    patch_cells, patch = self.fetch_corp_patch(patch_relative_path)
                    a = np.append(a, patch)
                    cells += patch_cells
            ap = a.reshape((1, 1, cells, 2, 2, 2, 3))
            gf.resequence_nexus_corp(ap)  # move from Nexus corp ordering to Pagoda ordering
            if (self.use_case, grid_name) in self.grid_extents_kji:
                extent_kji = self.grid_extents_kji[self.use_case, grid_name]
                assert cells == extent_kji[0] * extent_kji[1] * extent_kji[2], 'corp extent mismatch in vdb'
            else:
                extent_kji = gf.determine_corp_extent(ap)
                if extent_kji is None:
                    log.warning('failed to determine extent of root grid from corp data')
                    extent_kji = (1, 1, cells)
                else:
                    self.grid_extents_kji[self.use_case, grid_name] = extent_kji
            shape_7d = (extent_kji[0], extent_kji[1], extent_kji[2], 2, 2, 2, 3)
            return ap.reshape(shape_7d)
        except Exception:
            log.exception('failed to extract root grid corp data from vdb')
        return None

    def load_init_mapdata_array(self, file, keyword, dtype = None, unpack = False, grid_name = 'ROOT'):
        """Loads an INIT MAPDATA array from vdb; returns 3D numpy array coerced to dtype (if not None)."""

        try:
            if self.use_case is None:
                self.cases()
            assert self.use_case is not None, 'no case found in vdb'
            relative_path = os.path.join('INIT', 'MAPDATA', file)
            a = self.data_for_keyword(relative_path, keyword, search = True).a
            assert a is not None, 'failed to extract data for keyword ' + keyword + ' from vdb relative path ' + relative_path
            if unpack:
                un = self.grid_unpack(grid_name)
                if un is not None:
                    null_start = np.zeros((a.size + 1,), dtype = a.dtype)  # better to use Nan for null value?
                    null_start[1:] = a
                    a = null_start[un]
            a = self.grid_shaped(grid_name, coerce(a, dtype))
            return a
        except Exception:
            log.exception('failed to extract data from vdb for grid ' + str(grid_name))
        return None

    def load_recurrent_mapdata_array(self, file, keyword, dtype = None, unpack = False, grid_name = 'ROOT'):
        """Loads a RECUR MAPDATA array from vdb; returns 3D numpy array coerced to dtype (if not None)."""

        try:
            if bad_keyword(keyword):
                log.warning('ignoring attempt to load recurrent vdb mapdata for corrupt keyword')
                return None
            if self.use_case is None:
                self.cases()
            assert self.use_case is not None, 'no case found in vdb'
            relative_path = os.path.join('RECUR', 'MAPDATA', file)
            data = self.data_for_keyword(relative_path, keyword, search = True)
            if not data:
                return None
            a = data.a
            if a is None:
                return None
            if unpack:
                un = self.grid_unpack(grid_name)
                if un is not None:
                    null_start = np.zeros((a.size + 1,), dtype = a.dtype)  # better to use Nan for null value?
                    null_start[1:] = a
                    a = null_start[un]
            a = self.grid_shaped(grid_name, coerce(a, dtype))
            return a
        except Exception:
            log.exception('failed to extract grid recurrent data from vdb for keyword: ' + keyword)
        return None

    def root_dad(self):
        """Loads and returns the IROOTDAD array from vdb; returns 3D numpy int32 array."""

        return self.grid_dad('ROOT')

    def grid_dad(self, grid_name):
        """Loads and returns the DAD array from vdb for the named grid; returns 3D numpy int32 array."""

        # todo: check file naming and values for LGRs
        grid_name = grid_name.upper
        # DAD data appear to be unique cell ids, in usual sequence, starting at 1, same as UID
        return self.load_init_mapdata_array(file = 'I' + grid_name + 'DAD.bin',
                                            keyword = 'DAD',
                                            dtype = 'int32',
                                            unpack = False,
                                            grid_name = grid_name)

    def root_kid(self):
        """Loads and returns the IROOTKID array from vdb; returns 3D numpy int32 array (can be inactive cell mask)."""

        return self.grid_kid('ROOT')

    def grid_kid(self, grid_name):
        """Loads and returns the IROOTKID array from vdb; returns 3D numpy int32 array (can be inactive cell mask)."""

        grid_name = grid_name.upper()
        # see comments in grid_kid_inactive_mask() for KID values
        return self.load_init_mapdata_array(file = 'I' + grid_name + 'KID.bin',
                                            keyword = 'KID',
                                            dtype = 'int32',
                                            unpack = False,
                                            grid_name = grid_name)

    def root_kid_inactive_mask(self):
        """Loads the IROOTKID array and returns boolean mask of cells inactive in ROOT grid."""

        return self.grid_kid_inactive_mask('ROOT')

    def grid_kid_inactive_mask(self, grid_name):
        """Loads the KID array for the named grid and returns boolean mask of cells inactive in grid."""

        # KID values:
        # 0 - active in this grid
        # >0 - (in ROOT grid) cell inactive in ROOT as assigned to an LGR (KID value is LGR number)
        # -1 – inactive for normal reasons (zero pore volume or DEADCELL set)
        # -2 - (in LGR grids) cell omitted from LGR (possibly omitted as inactive in parent grid?)
        # -3 - inactive due to coarsening (coarsening repurposes one cell in parent grid for each coarsened cell)
        i = self.grid_kid(grid_name)
        if i is None:
            return None
        b = np.empty(i.shape, dtype = 'bool')
        b[:] = (i != 0)
        return b

    def root_uid(self):
        """Loads and returns the IROOTUID array from vdb; returns 3D numpy int32 array."""

        return self.grid_uid('ROOT')

    def grid_uid(self, grid_name):
        """Loads and returns the UID array from vdb for the named grid; returns 3D numpy int32 array."""

        grid_name = grid_name.upper()
        # UID data appear to be unique cell ids, in usual sequence, starting at 1, same as DAD
        return self.load_init_mapdata_array(file = 'I' + grid_name + 'UID.bin',
                                            keyword = 'UID',
                                            dtype = 'int32',
                                            unpack = False,
                                            grid_name = grid_name)

    def root_unpack(self):
        """Loads and returns the IROOTUNPACK array from vdb; returns 3D numpy int32 array."""

        return self.grid_unpack('ROOT')

    def grid_unpack(self, grid_name):
        """Loads and returns the IROOTUNPACK array from vdb; returns 3D numpy int32 array."""

        grid_name = grid_name.upper()
        if self.use_case is not None and (self.use_case, grid_name) in self.grid_packings.keys():
            return self.grid_packings[self.use_case, grid_name]
        # UNPACK data appear to be index into packed array for each cell, zero for inactive/absent
        un = self.load_init_mapdata_array(file = 'I' + grid_name + 'UNPACK.bin',
                                          keyword = 'UNPACK',
                                          dtype = 'int32',
                                          unpack = False,
                                          grid_name = grid_name)
        if self.use_case is not None:
            self.grid_packings[self.use_case,
                               grid_name] = un  # cache unpack array for later packing & unpacking operations
        return un

    def list_of_static_properties(self):
        """Returns list of static property keywords present in the vdb for ROOT."""

        return self.grid_list_of_static_properties('ROOT')

    def grid_list_of_static_properties(self, grid_name):
        """Returns list of static property keywords present in the vdb for named grid."""

        grid_name = grid_name.upper()
        glob_path = os.path.join(self.path, self.use_case, 'INIT', 'MAPDATA', 'I' + grid_name + '*.bin')
        if self.zipped:
            file_list = self.zip_glob(glob_path)
        else:
            file_list = glob.glob(glob_path)
        keyword_list = []
        for p in file_list:
            _, f = os.path.split(p)
            keyword = f[1 + len(grid_name):-4]
            keyword_list.append(keyword)
        return keyword_list

    def root_static_property(self, keyword, dtype = None, unpack = None):
        """Loads and returns a ROOT static property array."""

        return self.grid_static_property('ROOT'.keyword, dtype = dtype, unpack = unpack)

    def grid_static_property(self, grid_name, keyword, dtype = None, unpack = None):
        """Loads and returns a static property array for named grid."""

        grid_name = grid_name.upper()
        keyword = keyword.strip().upper()
        if unpack is None:
            unpack = (keyword not in init_not_packed)
        return self.load_init_mapdata_array(file = 'I' + grid_name + keyword + '.bin',
                                            keyword = keyword,
                                            dtype = dtype,
                                            unpack = unpack,
                                            grid_name = grid_name)

    def list_of_timesteps(self):
        """Returns a list of integer timesteps for which a ROOT recurrent mapdata file exists."""

        return self.grid_list_of_timesteps('ROOT')

    def grid_list_of_timesteps(self, grid_name):
        """Returns a list of integer timesteps for which a recurrent mapdata file for the named grid exists."""

        # todo: check file naming for LGR recurrent properties
        grid_name = grid_name.upper()
        glob_path = os.path.join(self.path, self.use_case, 'RECUR', 'MAPDATA', 'R' + grid_name + '_*.bin')
        if self.zipped:
            file_list = self.zip_glob(glob_path)
        else:
            file_list = glob.glob(glob_path)
        timestep_list = []
        for p in file_list:
            timestep_list.append(int(p[p.rfind('_') + 1:-4]))
        return sorted(timestep_list)

    def list_of_recurrent_properties(self, timestep):
        """Returns list of recurrent property keywords present in the vdb for given timestep."""

        return self.grid_list_of_recurrent_properties('ROOT', timestep)

    def grid_list_of_recurrent_properties(self, grid_name, timestep):
        """Returns list of recurrent property keywords present in the vdb for named grid for given timestep."""

        def keyword_list_from_fp(fp, path):
            """Returns a list of keywords (strings) found in a recurrent file."""
            assert fp.read(4) == b'NT32', 'first 4 characters not NT32 in file ' + path
            kp = KP(fp)
            keyword_list = kp.sub_key_list('MAPDATA', filter = True)
            try:
                keyword_list.remove('LASTMOD')
            except Exception:
                pass
            bad_key_indices = []
            for i in range(len(keyword_list)):
                key = keyword_list[i]
                if bad_keyword(key):
                    bad_key_indices.append(i)
            if len(bad_key_indices):
                log.warning(str(len(bad_key_indices)) + ' non-ascii keywords ignored in recurrent file ' + path)
                bad_key_indices.reverse()
                for i in bad_key_indices:
                    keyword_list.pop(i)
            return keyword_list

        # todo: check file naming for LGR recurrent properties
        grid_name = grid_name.upper()
        try:
            path = os.path.join(self.path, self.use_case, 'RECUR', 'MAPDATA',
                                'R' + grid_name + '_' + str(timestep) + '.bin')
            if self.zipped:
                with zf.ZipFile(self.zip_file) as zfp:
                    with zfp.open(path) as fp:
                        return keyword_list_from_fp(fp, path)
            else:
                assert os.path.exists(path), 'failed to find vdb file ' + str(path)
                with open(path, 'rb') as fp:
                    return keyword_list_from_fp(fp, path)
        except Exception:
            log.exception('failed to read keyword data from vdb binary file: ' + path)
        return None

    def root_recurrent_property_for_timestep(self, keyword, timestep, dtype = None, unpack = True):
        """Loads and returns a ROOT recurrent property array for one timestep."""

        return self.grid_recurrent_property_for_timestep('ROOT', keyword, timestep, dtype = dtype, unpack = unpack)

    def grid_recurrent_property_for_timestep(self, grid_name, keyword, timestep, dtype = None, unpack = True):
        """Loads and returns a recurrent property array for named grid for one timestep."""

        # todo: check file naming convention for LGRs
        grid_name = grid_name.upper()
        keyword = keyword.strip().upper()
        return self.load_recurrent_mapdata_array(file = 'R' + grid_name + '_' + str(timestep) + '.bin',
                                                 keyword = keyword,
                                                 dtype = dtype,
                                                 unpack = unpack,
                                                 grid_name = grid_name)

    def header_place_for_keyword(self, relative_path, keyword, search = True):
        """Low level function to return file position for header relating to given keyword."""

        try:
            path = os.path.join(self.path, relative_path)
            assert os.path.exists(path), 'failed to find vdb file ' + str(path)
            with open(path, 'rb') as fp:
                assert fp.read(4) == b'NT32', 'first 4 characters not NT32 in file ' + path
                kp = KP(fp)
                return kp.header_place_for_key(keyword, search = search)
        except Exception:
            log.exception('failed to read keyword data from vdb binary file: ' + path)
        return None

    def root_shaped(self, a):
        """Returns array reshaped to root grid extent for current use case, if known; otherwise unchanged."""

        return self.grid_shaped('ROOT', a)

    def grid_shaped(self, grid_name, a):
        """Returns array reshaped to named grid extent for current use case, if known; otherwise unchanged."""

        grid_name = grid_name.upper()
        if self.use_case is not None and (self.use_case, grid_name) in self.grid_extents_kji.keys():
            cell_count = np.prod(self.grid_extents_kji[self.use_case, grid_name])
            if a.size == cell_count:
                return a.reshape(tuple(self.grid_extents_kji[self.use_case, grid_name]))
        return a

    def zip_glob(self, path_with_asterisk):
        """Performs glob.glob like function for zipped file, path must contain a single asterisk."""

        assert self.zipped

        star = path_with_asterisk.find('*')
        assert star >= 0, 'no asterisk in zip glob path'
        no_tail = (star == len(path_with_asterisk) - 1)
        match_list = []
        with zf.ZipFile(self.zip_file) as zfp:
            zip_list = zfp.namelist()
            for name in zip_list:
                if ((star == 0 or name.startswith(path_with_asterisk[:star])) and
                    (no_tail or name.endswith(path_with_asterisk[star + 1:]))):
                    match_list.append(name)
        return match_list


def bad_keyword(key):
    """Return False if key is a valid keyword, otherwise True."""
    if not key:
        return True
    if key.isalnum():
        return False
    for c in key:
        if not c.isalnum() and c not in '-_':
            return True
    return False



"""Multiprocessing module containing the function used to run the wrapper functions in parallel."""

import logging
import os
import time
import uuid
from typing import List, Dict, Any, Callable, Union
from pathlib import Path
from joblib import Parallel, delayed, parallel_backend  # type: ignore

import resqpy.model as rq

log = logging.getLogger(__name__)


def rm_tree(path: Union[Path, str]) -> None:
    """Removes a directory using a pathlib Path.

    Args:
        path (Path/str): pathlib Path or string of the directory path.

    Returns:
        None.
    """
    path = Path(path)
    if os.path.isdir(path):
        for child in path.iterdir():
            if child.is_file():
                child.unlink()
            else:
                rm_tree(child)
        path.rmdir()


def function_multiprocessing(function: Callable,
                             kwargs_list: List[Dict[str, Any]],
                             recombined_epc: Union[Path, str],
                             cluster,
                             consolidate: bool = True,
                             require_success = False,
                             tmp_dir_path: Union[Path, str] = '.',
                             backend: str = 'dask') -> List[bool]:
    """Calls a function concurrently with the specfied arguments.

    A multiprocessing pool is used to call the function multiple times in parallel. Once
    all results are returned, they are combined into a single epc file.

    Args:
        function (Callable): the wrapper function to be called. Needs to return:

            - index (int): the index of the kwargs in the kwargs_list.
            - success (bool): whether the function call was successful, whatever that
                definiton is.
            - epc_file (Path/str): the epc file path where the objects are stored.
            - uuid_list (List[str]): list of UUIDs of relevant objects.

        kwargs_list (List[Dict[Any]]): A list of keyword argument dictionaries that are
            used when calling the function.
        recombined_epc (Path/str): A pathlib Path or path string of
            where the combined epc will be saved.
        cluster: if using the Dask backend, a LocalCluster is a Dask cluster on a
            local machine. If using a job queing system, a JobQueueCluster can be used
            such as an SGECluster, SLURMCluster, PBSCluster, LSFCluster etc.
        consolidate (bool): if True and an equivalent part already exists in
            a model, it is not duplicated and the uuids are noted as equivalent.
        require_success (bool): if True and any instance fails, then an exception is
            raised
        tmp_dir_path (str): path where the temporary directory is saved. Defaults to
            the calling code directory.
        backend (str): the joblib parallel backend used. Dask is used by default
            so a Dask cluster must be passed to the cluster argument.

    Returns:
        success_list (List[bool]): A boolean list of successful function calls.

    Note:
        This function uses the Dask backend by default to run the given function in
        parallel, so a Dask cluster must be setup and passed as an argument if Dask is
        used. Dask will need to be installed in the Python environment because it is not
        a dependency of the project. More info can be found at 
        https://resqpy.readthedocs.io/en/latest/tutorial/multiprocessing.html
    """
    log.info("multiprocessing function called with %s function, %s entries.", function.__name__, len(kwargs_list))

    tmp_dir = Path(tmp_dir_path) / f'tmp_{uuid.uuid4()}'
    for i, kwargs in enumerate(kwargs_list):
        kwargs["index"] = i
        kwargs["parent_tmp_dir"] = str(tmp_dir)

    if cluster is None:
        results = []
        for i, kwargs in enumerate(kwargs_list):
            log.debug(f'calling function for entry {i}; name: {kwargs.get("name")}')
            one_r = function(**kwargs)
            results.append(one_r)
            log.debug(f'completed entry: {one_r[0]}; success: {one_r[1]}; epc: {one_r[2]}')
            log.debug(f'uuid list: {one_r[3]}')
    else:
        with parallel_backend(backend):
            results = Parallel()(delayed(function)(**kwargs) for kwargs in kwargs_list)

    # Sorting the results by the original kwargs_list index.
    results = list(sorted(results, key = lambda x: x[0]))

    success_list = [result[1] for result in results]
    epc_list = [result[2] for result in results]
    uuids_list = [result[3] for result in results]
    success_count = sum(success_list)
    log.info("multiprocessing function calls complete; successes: %s/%s.", success_count, len(results))
    if require_success and success_count < len(results):
        raise Exception('one or more multiprocessing instances failed')

    epc_file = Path(str(recombined_epc))
    if epc_file.is_file():
        model_recombined = rq.Model(epc_file = str(epc_file), quiet = True)
        log.info(f"updating the recombined epc file: {epc_file}")
    else:
        model_recombined = rq.new_model(epc_file = str(epc_file))
        log.info(f"creating the recombined epc file: {epc_file}")

    model = None
    for i, epc in enumerate(epc_list):
        log.debug(f'recombining from mp instance {i} epc: {epc}')
        if epc is None:
            continue
        attempt = 0
        while not os.path.exists(epc):
            attempt += 1
            if attempt == 7:
                log.warning(f'mp epc slow to materialise: {epc}')
            if attempt > 300:
                raise FileNotFoundError(f'timeout waiting for multiprocess worker epc to become available: {epc}')
            time.sleep(min(attempt, 10))
        attempt = 0
        while True:
            attempt += 1
            try:
                model = rq.Model(epc_file = epc, quiet = True)
                break
            except FileNotFoundError:
                if attempt >= 10:
                    raise FileNotFoundError(f'timeout waiting for mp epc {epc}')
                time.sleep(1)
        uuids = uuids_list[i]
        if uuids is None:
            uuids = model.uuids()
        for u in uuids:
            attempt = 0
            while True:
                attempt += 1
                try:
                    model_recombined.copy_uuid_from_other_model(model, uuid = u, consolidate = consolidate)
                    break
                except BlockingIOError:
                    if attempt >= 5:
                        raise
                time.sleep(1)

    # Deleting temporary directory.
    log.debug(f"deleting the temporary directory {tmp_dir}")
    rm_tree(tmp_dir)

    model_recombined.store_epc(quiet = True)
    if model is not None:
        model.h5_release()

    log.debug(f"recombined epc file complete: {epc_file}")

    return success_list



"""read_nexus_fault.py: functions for reading Nexus fault definition data from an ascii file."""

# Nexus is a registered trademark of the Halliburton Company

import logging

log = logging.getLogger(__name__)

import os
import re
import unicodedata
import numpy as np
import pandas as pd


def load_nexus_fault_mult_table(file_name):
    """Reads a Nexus (!) format file containing one or more MULT keywords and returns a dataframe with the MULT rows."""

    def is_number(s):
        try:
            float(s)
            return True
        except ValueError:
            pass

        try:
            unicodedata.numeric(s)
            return True
        except (TypeError, ValueError):
            pass

        return False

    dfs = []

    num_tables = 0
    ISTABLE = False
    ISRECORD = False
    name = face = None
    grid = 'ROOT'  ## nexus default grid

    face_dict = {'TX': 'I', 'TY': 'J', 'TZ': 'K', 'TI': 'I', 'TJ': 'J', 'TK': 'K'}

    if os.path.isfile(file_name):
        chunks = []
        with open(file_name) as f:
            for line in f:
                if len(line.strip()):
                    if (not line.strip()[0] == '!') & (not line.strip()[0] == 'C'):
                        line = line.partition('!')[0]  # removing trailing comments
                        # line = line.partition('C')[0]  # removing trailing comments
                        tokens = line.split()
                        if ISTABLE:
                            if is_number(tokens[0]):
                                ISRECORD = True

                            if ISRECORD and (not is_number(tokens[0])):
                                data = chunks[0:]
                                d_elems = np.array([np.array(data[i].split()) for i in range(len(data))])
                                # fill empty elements with zero
                                lens = np.array([len(i) for i in d_elems])
                                # Mask of valid places in each row
                                mask = np.arange(lens.max()) < lens[:, None]
                                # Setup output array and put elements from data into masked positions
                                outdata = np.zeros(mask.shape, dtype = d_elems.dtype)
                                outdata[mask] = np.concatenate(d_elems)
                                df = pd.DataFrame(outdata)
                                for column in df.columns:
                                    df[column] = pd.to_numeric(df[column], errors = 'ignore')
                                df.columns = ['i1', 'i2', 'j1', 'j2', 'k1', 'k2', 'mult']
                                df['grid'] = grid
                                df['name'] = name
                                df['face'] = face
                                dfs.append(df)
                                num_tables += 1

                                ISTABLE = False
                                ISRECORD = False
                                chunks = []

                        if ISTABLE:
                            if re.match("(.*)GRID(.*)", tokens[0]):
                                if len(tokens) > 0:
                                    grid = tokens[1]
                            elif re.match("(.*)FNAME(.*)", tokens[0]):
                                if len(tokens) > 0:
                                    name = tokens[1]
                            else:
                                if re.match(r"^MULT$", tokens[0]):
                                    ISTABLE = False
                                    ISRECORD = False
                                    chunks = []
                                else:
                                    chunks.append(line.strip())

                        if re.match(r"^MULT$", tokens[0]):
                            if len(tokens) > 0:
                                face = face_dict[tokens[1]]
                            if 'MINUS' in tokens:
                                face += '-'  # indicates data apply to 'negative' faces of specified cells
                            grid = 'ROOT'  # nexus default
                            name = 'NONE'
                            ISTABLE = True

            else:
                if ISTABLE:
                    if ISRECORD:
                        data = chunks[0:]
                        d_elems = np.array([np.array(data[i].split()) for i in range(len(data))])
                        # fill empty elements with zero
                        lens = np.array([len(i) for i in d_elems])
                        # Mask of valid places in each row
                        mask = np.arange(lens.max()) < lens[:, None]
                        # Setup output array and put elements from data into masked positions
                        outdata = np.zeros(mask.shape, dtype = d_elems.dtype)
                        outdata[mask] = np.concatenate(d_elems)
                        df = pd.DataFrame(outdata)
                        for column in df.columns:
                            df[column] = pd.to_numeric(df[column], errors = 'ignore')
                        df.columns = ['i1', 'i2', 'j1', 'j2', 'k1', 'k2', 'mult']
                        df['grid'] = grid
                        df['name'] = name
                        df['face'] = face
                        dfs.append(df)
                        num_tables += 1

                        ISTABLE = False
                        ISRECORD = False
                        chunks = []

    fault_df = pd.concat(dfs).reset_index(drop = True)

    convert_dict = {'i1': int, 'i2': int, 'j1': int, 'j2': int, 'k1': int, 'k2': int, 'mult': float}
    fault_df = fault_df.astype(convert_dict)

    return fault_df


def load_nexus_fault_mult_table_new(file_name):
    """Reads a Nexus (!) format file containing one or more MULT keywords and returns a dataframe with the MULT rows."""

    return load_nexus_fault_mult_table(file_name)



"""Submodule containing the fault related grid functions."""

import logging

log = logging.getLogger(__name__)

import numpy as np

import resqpy.olio.vector_utilities as vec


def find_faults(grid, set_face_sets = False, create_organizing_objects_where_needed = False):
    """Searches for column-faces that are faulted and assigns fault ids; creates list of column-faces per fault id.

    note:
       this method is deprecated, or due for overhaul to make compatible with resqml_fault module and the
       GridConnectionSet class
    """

    # note:the logic to group kelp into distinct fault ids is simplistic and won't always give the right grouping

    if set_face_sets:
        grid.clear_face_sets()

    if hasattr(grid, 'fault_dict') and grid.fault_dict is not None and len(grid.fault_dict.keys()) > 0:
        if set_face_sets:
            for f, (j_list, i_list) in grid.fault_dict.items():
                fault_name = f'fault_{f}' if isinstance(f, int) else f
                grid.face_set_dict[fault_name] = (j_list, i_list, 'K')
            grid.set_face_set_gcs_list_from_dict(grid.face_set_dict, create_organizing_objects_where_needed)
        return None

    log.info('looking for faults in grid')
    grid.create_column_pillar_mapping()
    if not grid.has_split_coordinate_lines:
        log.info('grid does not have split coordinate lines, ie. is unfaulted')
        grid.fault_dict = None
        return None

    # note: if Ni or Nj is 1, the kelp array has zero size, but that seems to be handled okay
    kelp_j = np.zeros((grid.extent_kji[1] - 1, grid.extent_kji[2]), dtype = 'int')  # fault id between cols j, j+1
    kelp_i = np.zeros((grid.extent_kji[1], grid.extent_kji[2] - 1), dtype = 'int')  # fault id between cols i, i+1

    last_fault_id = 0

    # look for splits affecting j faces
    for j in range(grid.extent_kji[1] - 1):
        for i in range(grid.extent_kji[2]):
            if (grid.pillars_for_column[j, i, 1, 0] != grid.pillars_for_column[j + 1, i, 0, 0] or
                    grid.pillars_for_column[j, i, 1, 1] != grid.pillars_for_column[j + 1, i, 0, 1]):
                if i > 0 and kelp_j[j, i - 1] > 0:
                    kelp_j[j, i] = kelp_j[j, i - 1]
                else:
                    last_fault_id += 1
                    kelp_j[j, i] = last_fault_id

    # look for splits affecting i faces
    for i in range(grid.extent_kji[2] - 1):
        for j in range(grid.extent_kji[1]):
            if (grid.pillars_for_column[j, i, 0, 1] != grid.pillars_for_column[j, i + 1, 0, 0] or
                    grid.pillars_for_column[j, i, 1, 1] != grid.pillars_for_column[j, i + 1, 1, 0]):
                if j > 0 and kelp_i[j - 1, i] > 0:
                    kelp_i[j, i] = kelp_i[j - 1, i]
                else:
                    last_fault_id += 1
                    kelp_i[j, i] = last_fault_id

    # make pass over kelp to reduce distinct ids: combine where pillar has exactly 2 kelps, one in each of i and j
    if kelp_j.size and kelp_i.size:
        for j in range(grid.extent_kji[1] - 1):
            for i in range(grid.extent_kji[2] - 1):
                if (bool(kelp_j[j, i]) != bool(kelp_j[j, i + 1])) and (bool(kelp_i[j, i]) != bool(kelp_i[j + 1, i])):
                    j_id = kelp_j[j, i] + kelp_j[j, i + 1]  # ie. the non-zero value
                    i_id = kelp_i[j, i] + kelp_i[j + 1, i]
                    if j_id == i_id:
                        continue
                    # log.debug('merging fault id {} into {}'.format(i_id, j_id))
                    kelp_i = np.where(kelp_i == i_id, j_id, kelp_i)
                    kelp_j = np.where(kelp_j == i_id, j_id, kelp_j)

    fault_id_list = np.unique(np.concatenate(
        (np.unique(kelp_i.flatten()), np.unique(kelp_j.flatten()))))[1:]  # discard zero from list
    log.info('number of distinct faults: ' + str(fault_id_list.size))
    # for each fault id, make pair of tuples of kelp locations
    grid.fault_dict = {}  # maps fault_id to pair (j faces, i faces) of array of [j, i] kelp indices for that fault_id
    for fault_id in fault_id_list:
        grid.fault_dict[fault_id] = (np.stack(np.where(kelp_j == fault_id),
                                              axis = 1), np.stack(np.where(kelp_i == fault_id), axis = 1))
    grid.fault_id_j = kelp_j.copy()  # fault_id for each internal j kelp, zero is none; extent nj-1, ni
    grid.fault_id_i = kelp_i.copy()  # fault_id for each internal i kelp, zero is none; extent nj, ni-1
    if set_face_sets:
        for f, (j_list, i_list) in grid.fault_dict.items():
            fault_name = f'fault_{f}'
            grid.face_set_dict[fault_name] = (j_list, i_list, 'K')
        grid.set_face_set_gcs_list_from_dict(grid.face_set_dict, create_organizing_objects_where_needed)
    return (grid.fault_id_j, grid.fault_id_i)


def fault_throws(grid):
    """Finds mean throw of each J and I face; adds throw arrays as attributes to this grid and returns them.

    note:
       this method is deprecated, or due for overhaul to make compatible with resqml_fault module and the
       GridConnectionSet class
    """

    if hasattr(grid, 'fault_throw_j') and grid.fault_throw_j is not None and hasattr(
            grid, 'fault_throw_i') and grid.fault_throw_i is not None:
        return (grid.fault_throw_j, grid.fault_throw_i)
    if not grid.has_split_coordinate_lines:
        return None
    if not hasattr(grid, 'fault_id_j') or grid.fault_id_j is None or not hasattr(
            grid, 'fault_id_i') or grid.fault_id_i is None:
        grid.find_faults()
        if not hasattr(grid, 'fault_id_j'):
            return None
    log.debug('computing fault throws (deprecated method)')
    cp = grid.corner_points(cache_cp_array = True)
    grid.fault_throw_j = np.zeros((grid.nk, grid.nj - 1, grid.ni))
    grid.fault_throw_i = np.zeros((grid.nk, grid.nj, grid.ni - 1))
    grid.fault_throw_j = np.where(grid.fault_id_j == 0, 0.0,
                                  0.25 * np.sum(cp[:, 1:, :, :, 0, :, 2] - cp[:, :-1, :, :, 1, :, 2], axis = (3, 4)))
    grid.fault_throw_i = np.where(grid.fault_id_i == 0, 0.0,
                                  0.25 * np.sum(cp[:, :, 1:, :, :, 0, 2] - cp[:, :, -1:, :, :, 1, 2], axis = (3, 4)))
    return (grid.fault_throw_j, grid.fault_throw_i)


def fault_throws_per_edge_per_column(grid, mode = 'maximum', simple_z = False, axis_polarity_mode = True):
    """Return array holding max, mean or min throw based on split node separations.

    arguments:
       mode (string, default 'maximum'): one of 'minimum', 'mean', 'maximum'; determines how to resolve variation in throw for
          each column edge
       simple_z (boolean, default False): if True, the returned throw values are vertical offsets; if False, the displacement
          in xyz space between split points is the basis of the returned values and may include a lateral offset component as
          well as xy displacement due to sloping pillars
       axis_polarity (boolean, default True): determines shape and ordering of returned array; if True, the returned array has
          shape (nj, ni, 2, 2); if False the shape is (nj, ni, 4); see return value notes for more information

    returns:
       numpy float array of shape (nj, ni, 2, 2) or (nj, ni, 4) holding fault throw values for each column edge; units are
          z units of crs for this grid; if simple_z is False, xy units and z units must be the same; positive values indicate
          greater depth if z is increasing downwards (or shallower if z is increasing upwards); negative values indicate the
          opposite; the shape and ordering of the returned array is determined by axis_polarity_mode; if axis_polarity_mode is
          True, the returned array has shape (nj, ni, 2, 2) with the third index being axis (0 = J, 1 = I) and the final index
          being polarity (0 = minus face edge, 1 = plus face edge); if axis_polarity_mode is False, the shape is (nj, ni, 4)
          and the face edges are ordered I-, J+, I+, J-, as required by the resqml standard for a property with indexable
          element 'edges per column'

    notes:
       the throws calculated by this method are based merely on grid geometry and do not refer to grid connection sets;
       NB: the same absolute value is returned, with opposite sign, for the edges on opposing sides of a fault; either one of
       these alone indicates the full throw;
       the property module contains a pair of reformatting functions for moving an array between the two axis polarity modes;
       minimum and maximum modes work on the absolute throws
    """

    assert mode in ['maximum', 'mean', 'minimum']
    if not simple_z:
        assert grid.z_units() == grid.xy_units(), 'differing xy and z units not supported for non-simple-z fault throws'

    log.debug('computing fault throws per edge per column based on corner point geometry')
    if not grid.has_split_coordinate_lines:  # note: no NaNs returned in this situation
        if axis_polarity_mode:
            return np.zeros((grid.nj, grid.ni, 2, 2))
        return np.zeros((grid.nj, grid.ni, 4))
    grid.create_column_pillar_mapping()
    i_pillar_throws = (
        grid.points_cached[:, grid.pillars_for_column[:, :-1, :, 1], 2]
        -  # (nk+1, nj, ni-1, jp) +ve dz I- cell > I+ cell
        grid.points_cached[:, grid.pillars_for_column[:, 1:, :, 0], 2])
    j_pillar_throws = (grid.points_cached[:, grid.pillars_for_column[:-1, :, 1, :], 2] -
                       grid.points_cached[:, grid.pillars_for_column[1:, :, 0, :], 2])
    if not simple_z:
        i_pillar_throws = np.sign(
            i_pillar_throws)  # note: will return zero if displacement is purely horizontal wrt. z axis
        j_pillar_throws = np.sign(j_pillar_throws)
        i_pillar_throws *= vec.naive_lengths(grid.points_cached[:, grid.pillars_for_column[:, :-1, :, 1], :] -
                                             grid.points_cached[:, grid.pillars_for_column[:, 1:, :, 0], :])
        j_pillar_throws *= vec.naive_lengths(grid.points_cached[:, grid.pillars_for_column[:-1, :, 1, :], :] -
                                             grid.points_cached[:, grid.pillars_for_column[1:, :, 0, :], :])

    if mode == 'mean':
        i_edge_throws = np.nanmean(i_pillar_throws, axis = (0, -1))  # (nj, ni-1)
        j_edge_throws = np.nanmean(j_pillar_throws, axis = (0, -1))  # (nj-1, ni)
    else:
        min_i_edge_throws = np.nanmean(np.nanmin(i_pillar_throws, axis = 0), axis = -1)
        max_i_edge_throws = np.nanmean(np.nanmax(i_pillar_throws, axis = 0), axis = -1)
        min_j_edge_throws = np.nanmean(np.nanmin(j_pillar_throws, axis = 0), axis = -1)
        max_j_edge_throws = np.nanmean(np.nanmax(j_pillar_throws, axis = 0), axis = -1)
        i_flip_mask = (np.abs(min_i_edge_throws) > np.abs(max_i_edge_throws))
        j_flip_mask = (np.abs(min_j_edge_throws) > np.abs(max_j_edge_throws))
        if mode == 'maximum':
            i_edge_throws = np.where(i_flip_mask, min_i_edge_throws, max_i_edge_throws)
            j_edge_throws = np.where(j_flip_mask, min_j_edge_throws, max_j_edge_throws)
        elif mode == 'minimum':
            i_edge_throws = np.where(i_flip_mask, max_i_edge_throws, min_i_edge_throws)
            j_edge_throws = np.where(j_flip_mask, max_j_edge_throws, min_j_edge_throws)
        else:
            raise Exception('code failure')

    # positive values indicate column has greater z values, ie. downthrown if z increases with depth
    if axis_polarity_mode:
        throws = np.zeros((grid.nj, grid.ni, 2, 2))  # first 2 is I (0) or J (1); final 2 is -ve or +ve face
        throws[1:, :, 0, 0] = -j_edge_throws  # J-
        throws[:-1, :, 0, 1] = j_edge_throws  # J+
        throws[:, 1:, 1, 0] = -i_edge_throws  # I-
        throws[:, :-1, 1, 1] = i_edge_throws  # I+

    else:  # resqml protocol
        # order I-, J+, I+, J- as required for properties with 'edges per column' indexable element
        throws = np.zeros((grid.nj, grid.ni, 4))
        throws[:, 1:, 0] = -i_edge_throws  # I-
        throws[:-1, :, 1] = j_edge_throws  # J+
        throws[:, :-1, 2] = i_edge_throws  # I+
        throws[1:, :, 3] = -j_edge_throws  # J-

    return throws



"""Transmissibility functions for grids."""

# Nexus is a registered trademark of the Halliburton Company

import logging

log = logging.getLogger(__name__)

import numpy as np

import resqpy.fault as rqf
import resqpy.olio.vector_utilities as vec
import resqpy.olio.xml_et as rqet
import resqpy.organize as rqo


def half_cell_t(grid,
                perm_k = None,
                perm_j = None,
                perm_i = None,
                ntg = None,
                realization = None,
                darcy_constant = None,
                tolerance = 1.0e-6):
    """Creates a half cell transmissibilty property array for a regular or irregular IJK grid.

    arguments:
       grid (grid.Grid or grid.RegularGrid): the grid for which half cell transmissibilities are required
       perm_k, j, i (float arrays of shape (nk, nj, ni), optional): cell permeability values
          (for each direction), in mD; if None, the permeabilities are found in the grid's property collection
       ntg (float array of shape (nk, nj, ni), or float, optional): net to gross values to apply to I & J
          calculations; if a single float, is treated as a constant; if None, net to gross ratio data in the
          property collection is used
       realization (int, optional): if present and the property collection is scanned for perm or ntg
          arrays, only those properties for this realization will be used; ignored if arrays passed in
       darcy_constant (float, optional): if present, the value to use for the Darcy constant;
          if None, the grid's length units will determine the value as expected by Nexus
       tolerance (float, default 1.0e-6): minimum half axis length below which the transmissibility
          will be deemed uncomputable (for the axis in question); NaN values will be returned (not Inf);
          units are implicitly those of the grid's crs length units; ignored if grid is a RegularGrid

    returns:
       numpy float array of shape (nk, nj, ni, 3) if grid is a RegularGrid otherwise (nk, nj, ni, 3, 2)
          where the 3 covers K,J,I and (for irregular grids) the 2 covers the face polarity: - (0) and + (1);
          units will depend on the length units of the coordinate reference system for
          the grid (and implicitly on the units of the darcy_constant); if darcy_constant is None,
          the units will be m3.cP/(kPa.d) or bbl.cP/(psi.d) for grid length units of m and ft
          respectively

    notes:
       calls either for half_cell_t_irregular() or half_cell_t_regular() depending on class of grid;
       see also notes for half_cell_t_irregular() and half_cell_t_regular()
    """

    import resqpy.grid as grr

    if darcy_constant is None:
        length_units = grid.xy_units()
        assert length_units == 'm' or length_units.startswith('ft'), "Darcy constant must be specified"
        assert grid.z_units() == length_units
        darcy_constant = 0.008527 if length_units == 'm' else 0.001127  # these values from Nexus keyword ref.

    if perm_k is None or perm_j is None or perm_i is None or ntg is None:
        pc = grid.extract_property_collection()
        basic_5_parts = pc.basic_static_property_parts(realization = realization, share_perm_parts = True)
        if ntg is None and basic_5_parts[0] is not None:
            ntg = pc.cached_part_array_ref(basic_5_parts[0])
        if perm_k is None and basic_5_parts[4] is not None:
            perm_k = pc.cached_part_array_ref(basic_5_parts[4])
        if perm_j is None and basic_5_parts[3] is not None:
            perm_j = pc.cached_part_array_ref(basic_5_parts[3])
        if perm_i is None and basic_5_parts[2] is not None:
            perm_i = pc.cached_part_array_ref(basic_5_parts[2])

    assert perm_k is not None and perm_j is not None and perm_i is not None

    if isinstance(grid, grr.RegularGrid):
        return half_cell_t_regular(grid,
                                   perm_k = perm_k,
                                   perm_j = perm_j,
                                   perm_i = perm_i,
                                   ntg = ntg,
                                   darcy_constant = darcy_constant)
    elif isinstance(grid, grr.Grid):
        return half_cell_t_irregular(grid,
                                     perm_k = perm_k,
                                     perm_j = perm_j,
                                     perm_i = perm_i,
                                     ntg = ntg,
                                     darcy_constant = darcy_constant,
                                     tolerance = tolerance)
    else:
        raise ValueError(f'grid {type(grid)} is neither RegularGrid nor Grid object in call to half_cell_t()')


def half_cell_t_regular(grid, perm_k = None, perm_j = None, perm_i = None, ntg = None, darcy_constant = None):
    """Creates a half cell transmissibilty property array for a RegularGrid.

    arguments:
       grid (grid.RegularGrid): the grid for which half cell transmissibilities are required
       perm_k, j, i (float arrays of shape (nk, nj, ni), required): cell permeability values
          (for each direction), in mD;
       ntg (float array of shape (nk, nj, ni), or float, optional): net to gross values to apply to I & J
          calculations; if a single float, is treated as a constant; if None, a value of 1.0 is used
       darcy_constant (float, required): the value to use for the Darcy constant

    returns:
       numpy float array of shape (nk, nj, ni, 3) where the 3 covers K,J,I;
          units will depend on the length units of the coordinate reference system for
          the grid (and implicitly on the units of the darcy_constant); if darcy_constant is None,
          the units will be m3.cP/(kPa.d) or bbl.cP/(psi.d) for grid length units of m and ft
          respectively

    notes:
       the same half cell transmissibility value is applicable to both - and + polarity faces;
       the axes of the regular grid are assumed to be orthogonal;
       the net to gross factor is only applied to I & J transmissibilities, not K; the results
       include the Darcy Constant factor but not any transmissibility multiplier applied at the
       face; to compute the transmissibilty between neighbouring cells, take the harmonic mean of
       the two half cell transmissibilities and multiply by any transmissibility multiplier;
       returned array will need to be re-shaped before storing as a RESQML property
       with indexable elements of 'faces';
       the coordinate referemce system for the grid must have the same length units for xy and z;
       this function is vastly more computationally efficient than the general (irregular grid)
       function
    """

    assert perm_k is not None and perm_j is not None and perm_i is not None
    if ntg is None:
        ntg = 1.0

    axis_lengths = vec.naive_lengths(grid.block_dxyz_dkji)
    face_areas = np.array(
        (axis_lengths[1] * axis_lengths[2], axis_lengths[2] * axis_lengths[0], axis_lengths[0] * axis_lengths[1]))

    half_t = np.empty((grid.nk, grid.nj, grid.ni, 3))  # 3 is K,J,I
    half_t[..., 0] = perm_k * face_areas[0] / (0.5 * axis_lengths[0])
    half_t[..., 1] = ntg * perm_j * face_areas[1] / (0.5 * axis_lengths[1])
    half_t[..., 2] = ntg * perm_i * face_areas[2] / (0.5 * axis_lengths[2])

    return darcy_constant * half_t


def half_cell_t_irregular(grid,
                          perm_k = None,
                          perm_j = None,
                          perm_i = None,
                          ntg = None,
                          darcy_constant = None,
                          tolerance = 1.0e-6):
    """Creates a half cell transmissibilty property array for an IJK grid.

    arguments:
       grid (grid.Grid): the grid for which half cell transmissibilities are required
       perm_k, j, i (float arrays of shape (nk, nj, ni), reqwuired): cell permeability values
          (for each direction), in mD;
       ntg (float array of shape (nk, nj, ni), or float, required): net to gross values to apply to I & J
          calculations; if a single float, is treated as a constant;
       darcy_constant (float, required): the value to use for the Darcy constant
       tolerance (float, default 1.0e-6): minimum half axis length below which the transmissibility
          will be deemed uncomputable (for the axis in question); NaN values will be returned (not Inf);
          units are implicitly those of the grid's crs length units

    returns:
       numpy float array of shape (nk, nj, ni, 3, 2) where the 3 covers K,J,I and the 2 covers the
          face polarity: - (0) and + (1);
          units will depend on the length units of the coordinate reference system for
          the grid (and implicitly on the units of the darcy_constant); if darcy_constant is None,
          the units will be m3.cP/(kPa.d) or bbl.cP/(psi.d) for grid length units of m and ft
          respectively

    notes:
       the algorithm is equivalent to the half cell transmissibility element of the Nexus NEWTRAN
       calculation; each resulting value is effectively for the entire face, so area proportional
       fractions will be needed at faults with throw, or at grid boundaries; the net to gross
       factor is only applied to I & J transmissibilities, not K; the results
       include the Darcy Constant factor but not any transmissibility multiplier applied at the
       face; to compute the transmissibilty between neighbouring cells, take the harmonic mean of
       the two half cell transmissibilities and multiply by any transmissibility multiplier; if
       the two cells do not have simple sharing of a common face, first reduce each half cell
       transmissibility by the proportion of the face that is shared (which may be a different
       proportion for each of the two juxtaposed cells);
       returned array will need to be re-ordered and re-shaped before storing as a RESQML property
       with indexable elements of 'faces per cell';
       the coordinate referemce system for the grid must have the same length units for xy and z
    """

    # NB: axis argument is KJI index; this function also uses axes in xyz space

    assert perm_k is not None and perm_j is not None and perm_i is not None

    tolerance_sqr = tolerance * tolerance

    p = grid.points_ref(masked = False)
    edge_vectors = []
    pfc = None  # pillars for column mapping, for use when split pillars are present
    km = kp = None  # raw points k indices for K- and K+ faces, by layer, when K gaps present
    if grid.k_gaps:
        km = grid.k_raw_index_array
        kp = km + 1
    if grid.has_split_coordinate_lines:
        pfc = grid.create_column_pillar_mapping()
        if grid.k_gaps:
            edge_vectors.append(p[kp][:, pfc] - p[km][:, pfc])  # k edge vectors, shape (nk, nj, ni, 2, 2, 3)
            edge_vectors.append(
                p[:, pfc[:, :, 1, :]] -
                p[:, pfc[:, :, 0, :]])  # j edge vectors, shape (nk + k_gaps + 1, nj, ni, 2, 3) with jp removed
            edge_vectors.append(
                p[:, pfc[:, :, :, 1]] -
                p[:, pfc[:, :, :, 0]])  # i edge vectors, shape (nk + k_gaps + 1, nj, ni, 2, 3) with ip removed
        else:
            edge_vectors.append(p[1:, pfc] - p[:-1, pfc])  # k edge vectors, shape (nk, nj, ni, 2, 2, 3)
            edge_vectors.append(p[:, pfc[:, :, 1, :]] -
                                p[:, pfc[:, :, 0, :]])  # j edge vectors, shape (nk + 1, nj, ni, 2, 3) with jp removed
            edge_vectors.append(p[:, pfc[:, :, :, 1]] -
                                p[:, pfc[:, :, :, 0]])  # i edge vectors, shape (nk + 1, nj, ni, 2, 3) with ip removed
    else:
        if grid.k_gaps:
            edge_vectors.append(p[kp] - p[km])  # k edge vectors, shape (nk, nj + 1, ni + 1, 3)
            edge_vectors.append(p[:, 1:, :] - p[:, :-1, :])  # j edge vectors, shape (nk + k_gaps + 1, nj, ni + 1, 3)
            edge_vectors.append(p[:, :, 1:] - p[:, :, :-1])  # i edge vectors, shape (nk + k_gaps + 1, nj + 1, ni, 3)
        else:
            edge_vectors.append(p[1:] - p[:-1])  # k edge vectors, shape (nk, nj + 1, ni + 1, 3)
            edge_vectors.append(p[:, 1:] - p[:, :-1])  # j edge vectors, shape (nk + 1, nj, ni + 1, 3)
            edge_vectors.append(p[:, :, 1:] - p[:, :, :-1])  # i edge vectors, shape (nk + 1, nj + 1, ni, 3)

    half_t = np.empty((grid.nk, grid.nj, grid.ni, 3, 2))  # 3 is K,J,I; 2 is -/+ polarity (face)

    # note: for some reason half_axis_length_sqr values of 0.0 yield invalid value in numpy divide, rahter than divide by zero
    np.seterr(divide = 'ignore', invalid = 'ignore')

    for axis in range(3):

        if axis == 0:  # k
            if grid.has_split_coordinate_lines:
                half_axis_vectors = 0.125 * np.abs(np.sum(edge_vectors[0], axis = (
                    3, 4)))  # shape (nk, nj, ni, 3) with 3 being xyz length components of vector
                face_areas = (projected_tri_area(p[:, pfc[:, :, 0, 0]], p[:, pfc[:, :, 0, 1]], p[:, pfc[:, :, 1, 1]]) +
                              projected_tri_area(p[:, pfc[:, :, 0, 0]], p[:, pfc[:, :, 1, 1]], p[:, pfc[:, :, 1, 0]])
                             )  # shape (nk + k_gaps + 1, nj, ni, 3)
            else:
                half_axis_vectors = 0.125 * np.abs(edge_vectors[0][:, 1:, 1:] + edge_vectors[0][:, 1:, :-1] +
                                                   edge_vectors[0][:, :-1, 1:] + edge_vectors[0][:, :-1, :-1])
                face_areas = (projected_tri_area(p[:, :-1, :-1], p[:, :-1, 1:], p[:, 1:, 1:]) +
                              projected_tri_area(p[:, :-1, :-1], p[:, 1:, 1:], p[:, 1:, :-1])
                             )  # shape (nk + k_gaps + 1, nj, ni, 3) where 3 is xyz projection axis
            if grid.k_gaps:
                minus_face_areas = face_areas[
                    km]  # shape (nk, nj, ni, 3) where 3 is xyz projection axis (ie. yz plane, xz plane, xy plane)
                plus_face_areas = face_areas[kp]
            else:
                minus_face_areas = face_areas[:-1]
                plus_face_areas = face_areas[1:]
            half_axis_length_sqr = np.sum(half_axis_vectors * half_axis_vectors, axis = -1)  # shape (nk, nj, ni)
            zero_length_mask = np.logical_or(np.any(np.isnan(half_axis_vectors), axis = -1),
                                             half_axis_length_sqr < tolerance_sqr)
            minus_face_t = np.where(
                zero_length_mask, np.NaN,
                perm_k * np.sum(half_axis_vectors * minus_face_areas, axis = -1) / half_axis_length_sqr)
            plus_face_t = np.where(
                zero_length_mask, np.NaN,
                perm_k * np.sum(half_axis_vectors * plus_face_areas, axis = -1) / half_axis_length_sqr)

        elif axis == 1:  # j
            if grid.has_split_coordinate_lines:
                if grid.k_gaps:
                    half_axis_vectors = 0.125 * np.abs(edge_vectors[1][kp][:, :, :, 1] +
                                                       edge_vectors[1][kp][:, :, :, 0] +
                                                       edge_vectors[1][km][:, :, :, 1] +
                                                       edge_vectors[1][km][:, :, :, 0])
                    face_areas = (projected_tri_area(p[km][:, pfc[:, :, :, 0]], p[km][:, pfc[:, :, :, 1]],
                                                     p[kp][:, pfc[:, :, :, 1]]) +
                                  projected_tri_area(p[km][:, pfc[:, :, :, 0]], p[kp][:, pfc[:, :, :, 1]],
                                                     p[kp][:, pfc[:, :, :, 0]]))  # shape (nk, nj, ni, 2) where 2 is jp
                else:
                    half_axis_vectors = 0.125 * np.abs(edge_vectors[1][1:, :, :, 1] + edge_vectors[1][1:, :, :, 0] +
                                                       edge_vectors[1][:-1, :, :, 1] + edge_vectors[1][:-1, :, :, 0])
                    face_areas = (
                        projected_tri_area(p[:-1, pfc[:, :, :, 0]], p[:-1, pfc[:, :, :, 1]], p[1:, pfc[:, :, :, 1]]) +
                        projected_tri_area(p[:-1, pfc[:, :, :, 0]], p[1:, pfc[:, :, :, 1]], p[1:, pfc[:, :, :, 0]])
                    )  # shape (nk, nj, ni, 2) where 2 is jp
                minus_face_areas = face_areas[:, :, :, 0]
                plus_face_areas = face_areas[:, :, :, 1]
            else:
                if grid.k_gaps:
                    half_axis_vectors = 0.125 * np.abs(edge_vectors[1][kp][:, :, 1:] + edge_vectors[1][kp][:, :, :-1] +
                                                       edge_vectors[1][km][:, :, 1:] + edge_vectors[1][km][:, :, :-1])
                    face_areas = (projected_tri_area(p[km][:, :-1], p[km][:, :, 1:], p[kp][:, :, 1:]) +
                                  projected_tri_area(p[km][:, :-1], p[kp][:, :, 1:], p[kp][:, :, :-1]))
                else:
                    half_axis_vectors = 0.125 * np.abs(edge_vectors[1][1:, :, 1:] + edge_vectors[1][1:, :, :-1] +
                                                       edge_vectors[1][:-1, :, 1:] + edge_vectors[1][:-1, :, :-1])
                    face_areas = (projected_tri_area(p[:-1, :, :-1], p[:-1, :, 1:], p[1:, :, 1:]) +
                                  projected_tri_area(p[:-1, :, :-1], p[1:, :, 1:], p[1:, :, :-1])
                                 )  # shape (nk, nj + 1, ni)
                minus_face_areas = face_areas[:, :-1]
                plus_face_areas = face_areas[:, 1:]
            half_axis_length_sqr = np.sum(half_axis_vectors * half_axis_vectors, axis = -1)
            zero_length_mask = (half_axis_length_sqr < tolerance_sqr)
            minus_face_t = np.where(
                zero_length_mask, np.NaN,
                perm_j * np.sum(half_axis_vectors * minus_face_areas, axis = -1) / half_axis_length_sqr)
            plus_face_t = np.where(
                zero_length_mask, np.NaN,
                perm_j * np.sum(half_axis_vectors * plus_face_areas, axis = -1) / half_axis_length_sqr)

        else:  # i
            if grid.has_split_coordinate_lines:
                if grid.k_gaps:
                    half_axis_vectors = 0.125 * np.abs(edge_vectors[2][kp][:, :, :, 1] +
                                                       edge_vectors[2][kp][:, :, :, 0] +
                                                       edge_vectors[2][km][:, :, :, 1] +
                                                       edge_vectors[2][km][:, :, :, 0])
                    face_areas = (projected_tri_area(p[km][:, pfc[:, :, 0, :]], p[km][:, pfc[:, :, 1, :]],
                                                     p[kp][:, pfc[:, :, 1, :]]) +
                                  projected_tri_area(p[km][:, pfc[:, :, 0, :]], p[kp][:, pfc[:, :, 1, :]],
                                                     p[kp][:, pfc[:, :, 0, :]]))  # shape (nk, nj, ni, 2) where 2 is ip
                else:
                    half_axis_vectors = 0.125 * np.abs(edge_vectors[2][1:, :, :, 1] + edge_vectors[2][1:, :, :, 0] +
                                                       edge_vectors[2][:-1, :, :, 1] + edge_vectors[2][:-1, :, :, 0])
                    face_areas = (
                        projected_tri_area(p[:-1, pfc[:, :, 0, :]], p[:-1, pfc[:, :, 1, :]], p[1:, pfc[:, :, 1, :]]) +
                        projected_tri_area(p[:-1, pfc[:, :, 0, :]], p[1:, pfc[:, :, 1, :]], p[1:, pfc[:, :, 0, :]])
                    )  # shape (nk, nj, ni, 2) where 2 is ip
                minus_face_areas = face_areas[:, :, :, 0]
                plus_face_areas = face_areas[:, :, :, 1]
            else:
                if grid.k_gaps:
                    half_axis_vectors = 0.125 * np.abs(edge_vectors[2][kp][:, 1:, :] + edge_vectors[2][kp][:, :-1, :] +
                                                       edge_vectors[2][km][:, 1:, :] + edge_vectors[2][km][:, :-1, :])
                    face_areas = (projected_tri_area(p[km][:, :-1, :], p[km][:, 1:, :], p[kp][:, 1:, :]) +
                                  projected_tri_area(p[km][:, :-1, :], p[kp][:, 1:, :], p[kp][:, :-1, :]))
                else:
                    half_axis_vectors = 0.125 * np.abs(edge_vectors[2][1:, 1:, :] + edge_vectors[2][1:, :-1, :] +
                                                       edge_vectors[2][:-1, 1:, :] + edge_vectors[2][:-1, :-1, :])
                    face_areas = (projected_tri_area(p[:-1, :-1, :], p[:-1, 1:, :], p[1:, 1:, :]) +
                                  projected_tri_area(p[:-1, :-1, :], p[1:, 1:, :], p[1:, :-1, :]))
                minus_face_areas = face_areas[:, :, :-1]
                plus_face_areas = face_areas[:, :, 1:]
            half_axis_length_sqr = np.sum(half_axis_vectors * half_axis_vectors, axis = -1)
            zero_length_mask = (half_axis_length_sqr < tolerance_sqr)
            minus_face_t = np.where(
                zero_length_mask, np.NaN,
                perm_i * np.sum(half_axis_vectors * minus_face_areas, axis = -1) / half_axis_length_sqr)
            plus_face_t = np.where(
                zero_length_mask, np.NaN,
                perm_i * np.sum(half_axis_vectors * plus_face_areas, axis = -1) / half_axis_length_sqr)

        if axis != 0 and ntg is not None:
            minus_face_t *= ntg
            plus_face_t *= ntg

        half_t[:, :, :, axis, 0] = minus_face_t
        half_t[:, :, :, axis, 1] = plus_face_t

    np.seterr(divide = 'warn', invalid = 'warn')

    return np.abs(darcy_constant * half_t)


def half_cell_t_vertical_prism(vpg,
                               triple_perm_horizontal = None,
                               perm_k = None,
                               ntg = None,
                               darcy_constant = None,
                               tolerance = 1.0e-6):
    """Creates a half cell transmissibilty property array for a vertical prism grid.

    arguments:
       vpg (VerticalPrismGrid): the grid for which the half cell transmissibilities are required
       triple_perm_horizontal (numpy float array of shape (N, 3)): the directional permeabilities to apply
          to each of the three vertical faces per cell
       perm_k (numpy float array of shape (N,)): the permeability to use for the vertical transmissibilities
       ntg (numpy float array of shape (N,), optional): if present, acts as a multiplier in the
          computation of non-vertical transmissibilities
       darcy_constant (float, optional): the value to use for the Darcy constant; if None, a suitable
          value will be used depending on the length units of the vpg grid's crs
       tolerance (float, default 1.0e-6): minimum half axis length below which the transmissibility
          will be deemed uncomputable (for the axis in question); NaN values will be returned (not Inf);
          units are implicitly those of the grid's crs length units

    returns:
       numpy float array of shape (N, 5) being the per-face half cell transmissibilities for each cell

    note:
       order of 5 faces matches those of faces per cell, ie. top, base, then the 3 vertical faces
    """

    assert triple_perm_horizontal is not None
    assert triple_perm_horizontal.shape == (vpg.cell_count, 3)
    if perm_k is None:
        perm_k = np.mean(triple_perm_horizontal, axis = 1)
    if darcy_constant is None:
        length_units = vpg.xy_units()
        assert length_units == 'm' or length_units.startswith('ft')
        assert vpg.z_units() == length_units
        darcy_constant = 0.008527 if length_units == 'm' else 0.001127  # these values from Nexus keyword ref.

    # fetch triangulation and call precursor function
    p = vpg.points_ref()
    t = vpg.triangulation()
    a_t, d_t = half_cell_t_2d_triangular_precursor(p, t)
    # find heights of cells and faces
    # find horizontal area of triangles
    triangle_areas = vec.area_of_triangles(p, t, xy_projection = True).reshape((1, -1))
    cp = vpg.corner_points()
    half_thickness = 0.5 * vpg.thickness().reshape((vpg.nk, -1))

    # compute transmissibilities
    tr = np.zeros((vpg.cell_count, 5), dtype = float)
    # vertical
    tr[:, 0] = np.where(half_thickness < tolerance, np.NaN, (perm_k.reshape(
        (vpg.nk, -1)) * triangle_areas / half_thickness)).flatten()
    tr[:, 1] = tr[:, 0]
    # horizontal
    # TODO: compute dip adjustments for non-vertical transmissibilities
    dt_full = np.empty((vpg.nk, vpg.cell_count // vpg.nk, 3), dtype = float)
    dt_full[:] = d_t
    tr[:, 2:] = np.where(dt_full < tolerance, np.NaN,
                         triple_perm_horizontal.reshape((vpg.nk, -1, 3)) * a_t.reshape((1, -1, 3)) / dt_full).reshape(
                             (-1, 3))
    if ntg is not None:
        tr[:, 2:] *= ntg.reshape((-1, 1))

    tr *= darcy_constant
    return tr


def half_cell_t_2d_triangular_precursor(p, t):
    """Creates a precursor to horizontal transmissibility for prism grids (see notes).

    arguments:
       p (numpy float array of shape (N, 2 or 3)): the xy(&z) locations of cell vertices
       t (numpy int array of shape (M, 3)): the triangulation of p for which the transmissibility
          precursor is required

    returns:
       a pair of numpy float arrays, each of shape (M, 3) being the normal length and flow length
       relevant for flow across the face opposite each vertex as defined by t

    notes:
       this function acts as a precursor to the equivalent of the half cell transmissibility
       functions but for prism grids; for a resqpy VerticalPrismGrid, the triangulation can
       be shared by many layers with this function only needing to be called once; the first
       of the returned values (normal length) is the length of the triangle edge, in xy, when
       projected onto the normal of the flow direction; multiplying the normal length by a cell
       height will yield the area needed for transmissibility calculations; the second of the
       returned values (flow length) is the distance from the trangle centre to the midpoint of
       the edge and can be used as the distance term for a half cell transmissibilty; this
       function does not account for dip, it only handles the geometric aspects of half
       cell transmissibility in the xy plane
    """

    assert p.ndim == 2 and p.shape[1] in [2, 3]
    assert t.ndim == 2 and t.shape[1] == 3

    # centre points of triangles, in xy
    centres = np.mean(p[t], axis = 1)[:, :2]
    # midpoints of edges of triangles, in xy
    edge_midpoints = np.empty(tuple(list(t.shape) + [2]), dtype = float)
    edge_midpoints[:, 0, :] = 0.5 * (p[t[:, 1]] + p[t[:, 2]])[:, :2]
    edge_midpoints[:, 1, :] = 0.5 * (p[t[:, 2]] + p[t[:, 0]])[:, :2]
    edge_midpoints[:, 2, :] = 0.5 * (p[t[:, 0]] + p[t[:, 1]])[:, :2]
    # triangle edge vectors, projected in xy
    edge_vectors = np.empty(edge_midpoints.shape, dtype = float)
    edge_vectors[:, 0] = (p[t[:, 2]] - p[t[:, 1]])[:, :2]
    edge_vectors[:, 1] = (p[t[:, 0]] - p[t[:, 2]])[:, :2]
    edge_vectors[:, 2] = (p[t[:, 1]] - p[t[:, 0]])[:, :2]
    # vectors from triangle centres to mid points of edges (3 per triangle), in xy plane
    cem_vectors = edge_midpoints - centres.reshape((-1, 1, 2))
    cem_lengths = vec.naive_lengths(cem_vectors)
    # unit length vectors normal to cem_vectors, in the xy plane
    normal_vectors = np.zeros(edge_midpoints.shape)
    normal_vectors[:, :, 0] = cem_vectors[:, :, 1]
    normal_vectors[:, :, 1] = -cem_vectors[:, :, 0]
    normal_vectors = vec.unit_vectors(normal_vectors)
    # edge lengths projected onto normal vectors (length perpendicular to nominal flow direction)
    normal_lengths = np.abs(vec.dot_products(edge_vectors, normal_vectors))
    # return normal (cross-sectional) lengths and nominal flow direction lengths
    assert normal_lengths.shape == t.shape and cem_lengths.shape == t.shape

    return normal_lengths, cem_lengths


def fault_connection_set(grid, skip_inactive = False):
    """Builds a GridConnectionSet for juxtaposed faces where there is a split pillar, with fractional area data.

    arguments:
       grid (grid.Grid object): the grid for which a fault connection set is required
       skip_inactive (boolean, default False): if True, connections where either cell is inactive will be excluded

    returns:
       (GridConnectionSet, numpy float array of shape (count, 2)) where the connection set identifies all cell face
       pairs where there is juxtaposition and the array contains the fraction of the face areas that are juxtaposed;
       count is the number of cell face pairs in the connection set

    notes:
       the current algorithm is designed for faults where slip has occurred along pillars – sideways slip (strike-slip)
       will currently cause erroneous results; inaccuracies may also arise as pillars become less straight, less
       co-planar or less parallel; the combination of non-parallel pillars and layers of non-uniform thickness can
       produce inaccuracies in some situations; if the grid does not have split pillars (ie. is unfaulted), or if
       there are no qualifying connections across faults, then (None, None) will be returned;
       as fractional areas are returned, the results are applicable whether xy & z units are the same or differ
    """

    def all_nan(pillar):
        # return True if no valid points in pillar
        return np.all(np.any(np.isnan(pillar), axis = -1))

    def juxtapose(grid, p, pv, pam, pap, pbm, pbp, tol = 0.001):
        # return list of juxtaposed layer pairs and fractional areas: list of (ka, kb, fa, fb) and error info
        # TODO: orthogonal offset tolerance on pillars

        def pillar_flavour(pml_top, pml_bot, ppl_top, ppl_bot, tol = 0.001):
            # return flavour of points pattern on one pillar:
            # 1: both p above m top
            # 2: p top above m top, p bot within m
            # 3: p top above m top, p bot below m bot
            # 4: both p within m
            # 5: p top within m, p bot below m bot
            # 6: both p below m
            if ppl_top >= pml_top - tol and ppl_bot <= pml_bot + tol:
                return 4  # tolerance bias towards simplest flavour
            if ppl_bot <= pml_top + tol:
                return 1
            if ppl_top >= pml_bot - tol:
                return 6
            if ppl_top <= pml_top + tol:
                if ppl_bot <= pml_bot + tol:
                    return 2
                return 3
            return 5

        def basic_fr(g, h, ea, eb, tol):
            # returns fractional area for simple patterns (2 sides of overlap quadrilateral lie on pillars)
            if ea + eb < tol:
                return 0.0
            return (g + h) / (ea + eb)

        def fractional_area(paml_top,
                            paml_bot,
                            papl_top,
                            papl_bot,
                            pbml_top,
                            pbml_bot,
                            pbpl_top,
                            pbpl_bot,
                            tol = 0.001):
            # calculate fractional area of overlap, from m perspective (calling code swaps m & p for other perspective)
            fla = pillar_flavour(paml_top, paml_bot, papl_top, papl_bot, tol = tol)
            flb = pillar_flavour(pbml_top, pbml_bot, pbpl_top, pbpl_bot, tol = tol)
            if (fla, flb) == (4, 4):  # diagram 1
                return basic_fr(papl_bot - papl_top, pbpl_bot - pbpl_top, paml_bot - paml_top, pbml_bot - pbml_top, tol)
            elif (fla, flb) == (3, 3):  # diagram 1 (reverse perspective); diagram 3
                return 1.0
            elif (fla, flb) == (5, 5):  # diagram 2
                return basic_fr(paml_bot - papl_top, pbml_bot - pbpl_top, paml_bot - paml_top, pbml_bot - pbml_top, tol)
            elif (fla, flb) == (2, 2):  # diagram 2 (reverse perspective)
                return basic_fr(papl_bot - paml_top, pbpl_bot - pbml_top, paml_bot - paml_top, pbml_bot - pbml_top, tol)
            elif (fla, flb) == (5, 4):  # diagram 5 (diagram 4 is a special case of 5 and 2)
                ea = paml_bot - paml_top
                eb = pbml_bot - pbml_top
                g = pbml_bot - pbpl_bot
                if eb < tol or g < tol:
                    sub = 0.0
                else:
                    v = papl_bot - paml_bot
                    sub = (g / (ea + eb)) * (1.0 - v / (v + g))
                return basic_fr(paml_bot - papl_top, pbml_bot - pbpl_top, paml_bot - paml_top, pbml_bot - pbml_top,
                                tol) - sub
            elif (fla, flb) == (4, 5):  # diagram 5 mirror (diagram 4 is a special case of 5)
                ea = paml_bot - paml_top
                eb = pbml_bot - pbml_top
                g = paml_bot - papl_bot
                if ea < tol or g < tol:
                    sub = 0.0
                else:
                    v = pbpl_bot - pbml_bot
                    sub = (g / (ea + eb)) * (1.0 - v / (v + g))
                return basic_fr(pbml_bot - pbpl_top, paml_bot - papl_top, pbml_bot - pbml_top, paml_bot - paml_top,
                                tol) - sub
            elif (fla, flb) == (2, 3):  # diagram 5 (reverse perspective), similar to 1.0 - diagram 10
                ea = paml_bot - paml_top
                eb = pbml_bot - pbml_top
                g = paml_bot - papl_bot
                if ea < tol or g < tol:
                    return 1.0
                v = pbpl_bot - pbml_bot
                return 1.0 - (g / (ea + eb)) * (1.0 - v / (v + g))
            elif (fla, flb) == (3, 2):  # diagram 5 mirror (reverse perspective)
                ea = paml_bot - paml_top
                eb = pbml_bot - pbml_top
                g = pbml_bot - pbpl_bot
                if eb < tol or g < tol:
                    return 1.0
                v = papl_bot - paml_bot
                return 1.0 - (g / (ea + eb)) * (1.0 - v / (v + g))
            elif (fla, flb) == (5, 2):  # diagram 6
                ea = paml_bot - paml_top
                eb = pbml_bot - pbml_top
                g = papl_top - paml_top
                if ea < tol or g < tol:
                    suba = 0.0
                else:
                    v = pbml_top - pbpl_top
                    suba = (g / (ea + eb)) * (1.0 - v / (v + g))
                g = pbml_bot - pbpl_bot
                if eb < tol or g < tol:
                    subb = 0.0
                else:
                    v = papl_bot - paml_bot
                    subb = (g / (ea + eb)) * (1.0 - v / (v + g))
                sub = suba + subb
                return 1.0 - sub
            elif (fla, flb) == (2, 5):  # diagram 6 mirror
                ea = paml_bot - paml_top
                eb = pbml_bot - pbml_top
                g = pbpl_top - pbml_top
                if eb < tol or g < tol:
                    subb = 0.0
                else:
                    v = paml_top - papl_top
                    subb = (g / (ea + eb)) * (1.0 - v / (v + g))
                g = paml_bot - papl_bot
                if ea < tol or g < tol:
                    suba = 0.0
                else:
                    v = pbpl_bot - pbml_bot
                    suba = (g / (ea + eb)) * (1.0 - v / (v + g))
                sub = suba + subb
                return 1.0 - sub
            elif (fla, flb) == (2, 1):  # diagram 10
                ea = paml_bot - paml_top
                eb = pbml_bot - pbml_top
                g = papl_bot - paml_top
                if ea < tol or g < tol:
                    return 0.0
                v = pbml_top - pbpl_bot
                return (g / (ea + eb)) * (1.0 - v / (v + g))
            elif (fla, flb) == (1, 2):  # diagram 10 mirror
                eb = pbml_bot - pbml_top
                ea = paml_bot - paml_top
                g = pbpl_bot - pbml_top
                if eb < tol or g < tol:
                    return 0.0
                v = paml_top - papl_bot
                return (g / (eb + ea)) * (1.0 - v / (v + g))
            elif (fla, flb) == (5, 6):  # diagram 10 (reverse perspective)
                ea = paml_bot - paml_top
                eb = pbml_bot - pbml_top
                g = paml_bot - papl_top
                if ea < tol or g < tol:
                    return 0.0
                v = pbpl_top - pbml_bot
                return (g / (ea + eb)) * (1.0 - v / (v + g))
            elif (fla, flb) == (6, 5):  # diagram 10 mirror (reverse perspective)
                ea = paml_bot - paml_top
                eb = pbml_bot - pbml_top
                g = pbml_bot - pbpl_top
                if eb < tol or g < tol:
                    return 0.0
                v = papl_top - paml_bot
                return (g / (ea + eb)) * (1.0 - v / (v + g))
            elif (fla, flb) == (2, 4):  # diagram 11
                ea = paml_bot - paml_top
                eb = pbml_bot - pbml_top
                g = pbpl_top - pbml_top
                if eb < tol or g < tol:
                    sub = 0.0
                else:
                    v = paml_top - papl_top
                    sub = (g / (ea + eb)) * (1.0 - v / (v + g))
                return basic_fr(papl_bot - paml_top, pbpl_bot - pbml_top, paml_bot - paml_top, pbml_bot - pbml_top,
                                tol) - sub
            elif (fla, flb) == (4, 2):  # diagram 11 mirror
                ea = paml_bot - paml_top
                eb = pbml_bot - pbml_top
                g = papl_top - paml_top
                if ea < tol or g < tol:
                    sub = 0.0
                else:
                    v = pbml_top - pbpl_top
                    sub = (g / (ea + eb)) * (1.0 - v / (v + g))
                return basic_fr(pbpl_bot - pbml_top, papl_bot - paml_top, pbml_bot - pbml_top, paml_bot - paml_top,
                                tol) - sub
            elif (fla, flb) == (5, 3):  # diagram 11 (reverse perspective) similar to 1.0 - diagram 10
                ea = paml_bot - paml_top
                eb = pbml_bot - pbml_top
                g = papl_top - paml_top
                if ea < tol or g < tol:
                    return 1.0
                v = pbml_top - pbpl_top
                return 1.0 - (g / (ea + eb)) * (1.0 - v / (v + g))
            elif (fla, flb) == (3, 5):  # diagram 11 mirror (reverse perspective)
                ea = paml_bot - paml_top
                eb = pbml_bot - pbml_top
                g = pbpl_top - pbml_top
                if eb < tol or g < tol:
                    return 1.0
                v = paml_top - papl_top
                return 1.0 - (g / (ea + eb)) * (1.0 - v / (v + g))
            elif (fla, flb) == (3, 1):  # diagram 7 (only accurate if pillars parallel and layer constant thickness?)
                s = papl_bot - paml_bot
                ea = paml_bot - paml_top
                if s + ea <= tol:
                    return 0.0
                t = pbml_bot - pbpl_bot
                v = pbml_top - pbpl_bot
                if t + v <= tol:
                    return 1.0
                return 0.5 * (s / (s + t) + 1.0 - v / (v + ea + s))
            elif (fla, flb) == (1, 3):  # diagram 7 mirror
                # (only accurate if pillars parallel and layer constant thickness?)
                s = pbpl_bot - pbml_bot
                eb = pbml_bot - pbml_top
                if s + eb <= tol:
                    return 0.0
                t = paml_bot - papl_bot
                v = paml_top - papl_bot
                if t + v <= tol:
                    return 1.0
                return 0.5 * (s / (s + t) + 1.0 - v / (v + eb + s))
            elif (fla, flb) == (4, 6):  # diagram 7 (reverse perspective)
                ea = paml_bot - paml_top
                eb = pbml_bot - pbml_top
                g = paml_bot - papl_top
                if ea < tol or g < tol:
                    return 0.0
                v = pbpl_top - pbml_bot
                gs = paml_bot - papl_bot
                if gs < tol:
                    sub = 0.0
                else:
                    vs = pbpl_bot - pbml_bot
                    sub = (gs / (ea + eb)) * (1.0 - vs / (vs + gs))
                return (g / (ea + eb)) * (1.0 - v / (v + g)) - sub
            elif (fla, flb) == (6, 4):  # diagram 7 mirror (reverse perspective)
                ea = paml_bot - paml_top
                eb = pbml_bot - pbml_top
                g = pbml_bot - pbpl_top
                if eb < tol or g < tol:
                    return 0.0
                v = papl_top - paml_bot
                gs = pbml_bot - pbpl_bot
                if gs < tol:
                    sub = 0.0
                else:
                    vs = papl_bot - paml_bot
                    sub = (gs / (ea + eb)) * (1.0 - vs / (vs + gs))
                return (g / (ea + eb)) * (1.0 - v / (v + g)) - sub
            elif (fla, flb) == (4, 1):  # diagram 12
                ea = paml_bot - paml_top
                eb = pbml_bot - pbml_top
                g = papl_bot - paml_top
                if ea < tol or g < tol:
                    return 0.0
                gs = papl_top - paml_top
                if gs < tol:
                    sub = 0.0
                else:
                    v = pbml_top - pbpl_top
                    sub = (gs / (ea + eb)) * (1.0 - v / (v + gs))
                v = pbml_top - pbpl_bot
                return (g / (ea + eb)) * (1.0 - v / (v + g)) - sub
            elif (fla, flb) == (1, 4):  # diagram 12 mirror
                ea = paml_bot - paml_top
                eb = pbml_bot - pbml_top
                g = pbpl_bot - pbml_top
                if eb < tol or g < tol:
                    return 0.0
                gs = pbpl_top - pbml_top
                if gs < tol:
                    sub = 0.0
                else:
                    v = paml_top - papl_top
                    sub = (gs / (ea + eb)) * (1.0 - v / (v + gs))
                v = paml_top - papl_bot
                return (g / (ea + eb)) * (1.0 - v / (v + g)) - sub
            elif (fla, flb) == (3, 6):  # diagram 12 (reverse perspective)
                s = pbpl_top - pbml_bot
                eb = pbml_bot - pbml_top
                if s + eb <= tol:
                    return 1.0
                t = paml_bot - papl_top
                v = paml_top - papl_top
                if t + v <= tol:
                    return 0.0
                return 1.0 - (0.5 * (s / (s + t) + 1.0 - v / (v + eb + s)))
            elif (fla, flb) == (6, 3):  # diagram 12 mirror (reverse perspective)
                s = papl_top - paml_bot
                ea = paml_bot - paml_top
                if s + ea <= tol:
                    return 1.0
                t = pbml_bot - pbpl_top
                v = pbml_top - pbpl_top
                if t + v <= tol:
                    return 0.0
                return 1.0 - (0.5 * (s / (s + t) + 1.0 - v / (v + ea + s)))
            elif (fla, flb) == (5, 1):  # diagram 9 (only accurate if pillars parallel and layer constant thickness?)
                ea = paml_bot - paml_top
                eb = pbml_bot - pbml_top
                g = papl_top - paml_top
                if ea < tol or g < tol:
                    sub = 0.0
                else:
                    v = pbml_top - pbpl_top
                    sub = (g / (ea + eb)) * (1.0 - v / (v + g))
                s = papl_bot - paml_bot
                if s + ea <= tol:
                    return 0.0
                t = pbml_bot - pbpl_bot
                v = pbml_top - pbpl_bot
                if t + v <= tol:
                    return 1.0 - sub
                return 0.5 * (s / (s + t) + 1.0 - v / (v + ea + s)) - sub
            elif (fla, flb) == (1, 5):  # diagram 9 mirror
                ea = paml_bot - paml_top
                eb = pbml_bot - pbml_top
                g = pbpl_top - pbml_top
                if eb < tol or g < tol:
                    sub = 0.0
                else:
                    v = paml_top - papl_top
                    sub = (g / (ea + eb)) * (1.0 - v / (v + g))
                s = pbpl_bot - pbml_bot
                if s + eb <= tol:
                    return 0.0
                t = paml_bot - papl_bot
                v = paml_top - papl_bot
                if t + v <= tol:
                    return 1.0 - sub
                return 0.5 * (s / (s + t) + 1.0 - v / (v + eb + s)) - sub
            elif (fla, flb) == (2, 6):  # diagram 9 (reverse perspective)
                ea = paml_bot - paml_top
                eb = pbml_bot - pbml_top
                g = paml_bot - papl_bot
                if ea < tol or g < tol:
                    sub = 0.0
                else:
                    v = pbpl_bot - pbml_bot
                    sub = (g / (ea + eb)) * (1.0 - v / (v + g))
                s = paml_top - papl_top
                if s + ea <= tol:
                    return 0.0
                t = pbpl_top - pbml_top
                v = pbpl_top - pbml_bot
                if t + v <= tol:
                    return 1.0 - sub
                return 0.5 * (s / (s + t) + 1.0 - v / (v + ea + s)) - sub
            elif (fla, flb) == (6, 2):  # diagram 9 mirror (reverse perspective)
                ea = paml_bot - paml_top
                eb = pbml_bot - pbml_top
                g = pbml_bot - pbpl_bot
                if eb < tol or g < tol:
                    sub = 0.0
                else:
                    v = papl_bot - paml_bot
                    sub = (g / (ea + eb)) * (1.0 - v / (v + g))
                s = pbml_top - pbpl_top
                if s + eb <= tol:
                    return 0.0
                t = papl_top - paml_top
                v = papl_top - paml_bot
                if t + v <= tol:
                    return 1.0 - sub
                return 0.5 * (s / (s + t) + 1.0 - v / (v + eb + s)) - sub
            elif (fla, flb) == (6, 1):  # diagram 8; solution only accurate if pillars are parallel
                ea = paml_bot - paml_top
                eb = pbml_bot - pbml_top
                s = papl_top - paml_bot
                if s + ea <= tol:
                    suba = 0.0
                else:
                    t = pbml_bot - pbpl_top
                    v = pbml_top - pbpl_top
                    suba = 0.5 * (s / (s + t) + 1.0 - v / (v + ea + s)) if t + v > tol else 1.0
                s = pbml_top - pbpl_bot
                if s + eb <= tol:
                    subb = 0.0
                else:
                    t = papl_bot - paml_top
                    v = papl_bot - paml_bot
                    subb = 0.5 * (s / (s + t) + 1.0 - v / (v + eb + s)) if t + v > tol else 1.0
                sub = suba + subb
                return max(1.0 - sub, 0.0)
            elif (fla, flb) == (1, 6):  # diagram 8 mirror or reverse perspective
                ea = paml_bot - paml_top
                eb = pbml_bot - pbml_top
                s = pbpl_top - pbml_bot
                if s + eb <= tol:
                    subb = 0.0
                else:
                    t = paml_bot - papl_top
                    v = paml_top - papl_top
                    subb = 0.5 * (s / (s + t) + 1.0 - v / (v + eb + s)) if t + v > tol else 1.0
                s = paml_top - papl_bot
                if s + ea <= tol:
                    suba = 0.0
                else:
                    t = pbpl_bot - pbml_top
                    v = pbpl_bot - pbml_bot
                    suba = 0.5 * (s / (s + t) + 1.0 - v / (v + ea + s)) if t + v > tol else 1.0
                sub = suba + subb
                return max(1.0 - sub, 0.0)
            elif (fla, flb) == (4, 3):  # diagram 13
                ea = paml_bot - paml_top
                eb = pbml_bot - pbml_top
                if ea < tol:
                    return 1.0
                g = papl_top - paml_top
                if g < tol:
                    sub1 = 0.0
                else:
                    v = pbml_top - pbpl_top
                    sub1 = (g / (ea + eb)) * (1.0 - v / (v + g))
                g = paml_bot - papl_bot
                if g < tol:
                    sub2 = 0.0
                else:
                    v = pbpl_bot - pbml_bot
                    sub2 = (g / (ea + eb)) * (1.0 - v / (v + g))
                return 1.0 - (sub1 + sub2)
            elif (fla, flb) == (3, 4):  # diagram 13 mirror or reverse perspective
                ea = paml_bot - paml_top
                eb = pbml_bot - pbml_top
                if eb < tol:
                    return 1.0
                g = pbpl_top - pbml_top
                if g < tol:
                    sub1 = 0.0
                else:
                    v = paml_top - papl_top
                    sub1 = (g / (ea + eb)) * (1.0 - v / (v + g))
                g = pbml_bot - pbpl_bot
                if g < tol:
                    sub2 = 0.0
                else:
                    v = papl_bot - paml_bot
                    sub2 = (g / (ea + eb)) * (1.0 - v / (v + g))
                return 1.0 - (sub1 + sub2)
            else:
                raise Exception(f'unexpected juxtaposition pattern ({fla}, {flb})')

        pav = 0.5 * (pv[pam] + pv[pap])  # mean pillar unit vector for -I side of J face
        pbv = 0.5 * (pv[pbm] + pv[pbp])  # mean pillar unit vector for +I side of J face
        if all_nan(p[:, pam, :]) or all_nan(p[:, pap, :]) or all_nan(p[:, pbm, :]) or all_nan(p[:, pbp, :]):
            return []
        o = np.nanmin(p[:, pam, :], axis = 0)  # arbitrary local origin
        oe = np.expand_dims(o, 0)
        paml = p[:, pam, :] - oe  # pillar points translated to local space
        papl = p[:, pap, :] - oe
        pbml = p[:, pbm, :] - oe
        pbpl = p[:, pbp, :] - oe
        pamln = np.empty(
            paml.shape[0])  # pillar points normalised to scalar distances on mean pillar vector, arbitrary origin
        for i in range(pamln.size):
            pamln[i] = np.dot(paml[i], pav)
        papln = np.empty(papl.shape[0])
        for i in range(papln.size):
            papln[i] = np.dot(papl[i], pav)
        pbmln = np.empty(pbml.shape[0])
        for i in range(pbmln.size):
            pbmln[i] = np.dot(pbml[i], pbv)
        pbpln = np.empty(pbpl.shape[0])
        for i in range(pbpln.size):
            pbpln[i] = np.dot(pbpl[i], pbv)

        juxta_list = []
        fa_p_totals = np.zeros(grid.nk)
        fa_m_worst_scaling = 1.0
        fa_m_downscaling_count = 0
        for km in range(grid.nk):  # for each layer on -ve side of fault
            km_top = grid.k_raw_index_array[km] if grid.k_gaps else km  # index of top points on -ve side of fault
            km_bot = km_top + 1
            paml_top = pamln[
                km_top]  # point distances in local vector space; note local vector is different for pillars a and b
            paml_bot = pamln[km_bot]
            pbml_top = pbmln[km_top]
            pbml_bot = pbmln[km_bot]
            if np.any(np.isnan((paml_top, paml_bot, pbml_top, pbml_bot))):
                continue
            # scan layers on +ve side of fault looking for juxtaposition
            kp = -1
            km_start_index = len(juxta_list)
            fa_m_total = 0.0
            while True:
                kp += 1
                if kp >= grid.nk:
                    break
                kp_top = grid.k_raw_index_array[kp] if grid.k_gaps else kp  # index of top points on +ve side of fault
                kp_bot = kp_top + 1
                papl_top = papln[
                    kp_top]  # point distances in local vector space; note local vector is different for pillars a and b
                papl_bot = papln[kp_bot]
                pbpl_top = pbpln[kp_top]
                pbpl_bot = pbpln[kp_bot]
                if np.any(np.isnan((papl_top, papl_bot, pbpl_top, pbpl_bot))):
                    continue
                # in following comments, 'shallower' means less distance in local vector, which is a K direction vector of sorts
                if (paml_top >= papl_bot - tol) and (pbml_top >= pbpl_bot - tol):
                    continue  # p fully shallower than m
                if (paml_bot <= papl_top + tol) and (pbml_bot <= pbpl_top + tol):
                    continue  # p fully deeper than m
                # juxtaposition established, now determine fractional overlap area from perspectives of both sides of fault
                fa_m = fractional_area(paml_top,
                                       paml_bot,
                                       papl_top,
                                       papl_bot,
                                       pbml_top,
                                       pbml_bot,
                                       pbpl_top,
                                       pbpl_bot,
                                       tol = tol)
                fa_p = fractional_area(papl_top,
                                       papl_bot,
                                       paml_top,
                                       paml_bot,
                                       pbpl_top,
                                       pbpl_bot,
                                       pbml_top,
                                       pbml_bot,
                                       tol = tol)
                # log.debug(f'K {km} {fa_m}  <->  {fa_p} {kp}')
                fa_m = min(max(fa_m, 0.0), 1.0)
                fa_p = min(max(fa_p, 0.0), 1.0)
                juxta_list.append((km, kp, fa_m, fa_p))
                fa_m_total += fa_m
                fa_p_totals[kp] += fa_p
            # find sum of fractional areas by layer (separately for both perspectives); normalise to max of 1.0
            if fa_m_total > 1.0:
                fa_m_downscaling_count += 1
                if fa_m_total > fa_m_worst_scaling:
                    fa_m_worst_scaling = fa_m_total
                    # log.warning(f'downscaling fractional areas on minus side of fault by factor of {fa_m_total} in layer {km}')
                for i in range(km_start_index, len(juxta_list)):
                    (km, kp, fa_m, fa_p) = juxta_list[i]
                    juxta_list[i] = (km, kp, fa_m / fa_m_total, fa_p)
        any_p_scaling = False
        fa_p_worst_scaling = np.nanmax(fa_p_totals)
        fa_p_downscaling_count = 0
        for kp in range(grid.nk):
            if fa_p_totals[kp] > 1.0:
                fa_p_downscaling_count += 1
                # log.warning(f'downscaling fractional areas on plus side of fault by factor of {fa_p_totals[kp]} in layer {kp}')
                any_p_scaling = True
        if any_p_scaling:
            for i in range(len(juxta_list)):
                (km, kp, fa_m, fa_p) = juxta_list[i]
                if fa_p_totals[kp] > 1.0:
                    juxta_list[i] = (km, kp, fa_m, fa_p / fa_p_totals[kp])
        return juxta_list, (fa_m_downscaling_count, fa_p_downscaling_count, fa_m_worst_scaling, fa_p_worst_scaling)

    if not grid.has_split_coordinate_lines:
        return None, None
    skip_inactive = skip_inactive and hasattr(grid, 'inactive') and grid.inactive is not None

    p = grid.points_ref(masked = False)  # shape (nk + k_gaps + 1, np, 3)
    pv = vec.unit_vectors(p[-1] - p[0])  # pillar vectors; shape (np, 3); could postpone to individual pillar work
    col_j_split, col_i_split = grid.split_column_faces(
    )  # internal only column faces; shapes (nj - 1, ni), (nj, ni - 1)
    pfc = grid.create_column_pillar_mapping()  # pillars for column; shape (nj, ni, 2, 2)

    juxtaposed_j_list = []
    j_ji = np.stack(np.where(col_j_split)).T  # ji0 columns where +J face is split; shape (nc, 2)
    warning_count = 0
    for (j, i) in j_ji:
        # log.debug(f'J, I  {j}, {i}')
        pam = pfc[j, i, 1, 0]  # pillar index for -I edge of J face, for col on -J side of fault
        pap = pfc[j + 1, i, 0, 0]  # pillar index for -I edge of J face, for col on +J side of fault
        pbm = pfc[j, i, 1, 1]  # pillar index for +I edge of J face, for col on -J side of fault
        pbp = pfc[j + 1, i, 0, 1]  # pillar index for +I edge of J face, for col on +J side of fault
        ji_list, scaling_info = juxtapose(grid, p, pv, pam, pap, pbm, pbp)
        if scaling_info[2] > 1.2 or scaling_info[3] > 1.2:
            if warning_count < 20:
                log.warning(
                    f'severe downscaling for I+ face in column ({j}, {i}); worst m {scaling_info[2]}; worst p {scaling_info[3]}'
                )
            elif warning_count == 20:
                log.warning('other similar I face warnings suppressed')
            warning_count += 1
        for (km, kp, fa_m, fa_p) in ji_list:
            if skip_inactive and (grid.inactive[km, j, i] or grid.inactive[kp, j + 1, i]):
                continue
            juxtaposed_j_list.append(((km, j, i), (kp, j + 1, i), fa_m, fa_p))
    juxtaposed_i_list = []
    i_ji = np.stack(np.where(col_i_split)).T  # ji0 columns where +I face is split; shape (nc, 2)
    warning_count = 0
    for (j, i) in i_ji:
        pam = pfc[j, i, 0, 1]  # pillar index for -J edge of I face, for col on -I side of fault
        pap = pfc[j, i + 1, 0, 0]  # pillar index for -J edge of I face, for col on +I side of fault
        pbm = pfc[j, i, 1, 1]  # pillar index for +J edge of I face, for col on -I side of fault
        pbp = pfc[j, i + 1, 1, 0]  # pillar index for +J edge of I face, for col on +I side of fault
        ji_list, scaling_info = juxtapose(grid, p, pv, pam, pap, pbm, pbp)
        if scaling_info[2] > 1.2 or scaling_info[3] > 1.2:
            if warning_count < 20:
                log.warning(
                    f'severe downscaling for J+ face in column ({j}, {i}); worst m {scaling_info[2]}; worst p {scaling_info[3]}'
                )
            elif warning_count == 20:
                log.warning('other similar J face warnings suppressed')
            warning_count += 1
        for (km, kp, fa_m, fa_p) in ji_list:
            if skip_inactive and (grid.inactive[km, j, i] or grid.inactive[kp, j, i + 1]):
                continue
            juxtaposed_i_list.append(((km, j, i), (kp, j, i + 1), fa_m, fa_p))

    combo_list = juxtaposed_j_list + juxtaposed_i_list
    count = len(combo_list)
    if count == 0:
        return None, None

    # build connection set
    fcs = rqf.GridConnectionSet(grid.model, grid = grid)
    fcs.grid_list = [grid]
    fcs.count = count
    fcs.grid_index_pairs = np.zeros((count, 2), dtype = int)
    fcs.cell_index_pairs = np.zeros((count, 2), dtype = int)
    fcs.cell_index_pairs[:] = np.array([
        grid.natural_cell_indices(np.array([[k, j, i] for ((k, j, i), _, _, _) in combo_list])),
        grid.natural_cell_indices(np.array([[k, j, i] for (_, (k, j, i), _, _) in combo_list]))
    ]).T
    fcs.face_index_pairs = np.zeros((count, 2), dtype = int)
    fcs.face_index_pairs[:len(juxtaposed_j_list), 0] = fcs.face_index_map[1, 1]  # J+
    fcs.face_index_pairs[:len(juxtaposed_j_list), 1] = fcs.face_index_map[1, 0]  # J-
    fcs.face_index_pairs[len(juxtaposed_j_list):, 0] = fcs.face_index_map[2, 1]  # I+
    fcs.face_index_pairs[len(juxtaposed_j_list):, 1] = fcs.face_index_map[2, 0]  # I-

    feature_name = 'all faults with throw'
    fcs.feature_indices = np.zeros(count, dtype = int)  # could create seperate features by named fracture
    tbf = rqo.TectonicBoundaryFeature(grid.model, kind = 'fault', feature_name = feature_name)
    tbf_root = tbf.create_xml()
    fi = rqo.FaultInterpretation(grid.model, tectonic_boundary_feature = tbf, is_normal = True)
    fi_root = fi.create_xml(tbf_root, title_suffix = None)
    fi_uuid = rqet.uuid_for_part_root(fi_root)

    fcs.feature_list = [('obj_FaultInterpretation', fi_uuid, str(feature_name))]

    fa = np.array([[fa_m, fa_p] for (_, _, fa_m, fa_p) in combo_list])

    return fcs, fa


def projected_tri_area(pa, pb, pc):
    """Return array holding areas of triangles projected onto each of yz, xz, xy.

    arguments:
       pa, pb, pc (numpy float array of shape (..., 3): the corner points of a set of triangles (one corner
          in each of pa, pb & pc, for every triangle); last axis in arrays covers x,y,z

    returns:
       numpy float array of same shape as pa, pb & pc, with the last axis covering yz, xz, xy projections;
       the return values are the areas of the triangles projected in the three principal x,y,z axes

    note:
       assumes that units for x, y & z are the same; returned area units are those implicit units squared
    """

    assert pa.shape == pb.shape == pc.shape and pa.shape[-1] == 3

    area = np.empty(pa.shape)
    for xyz in range(3):
        xyz_0 = (xyz + 1) % 3
        xyz_1 = (xyz + 2) % 3
        pap = np.stack((pa[..., xyz_0], pa[..., xyz_1]), axis = -1)  # points projected onto 2D
        pbp = np.stack((pb[..., xyz_0], pb[..., xyz_1]), axis = -1)
        pcp = np.stack((pc[..., xyz_0], pc[..., xyz_1]), axis = -1)
        lap = vec.naive_lengths(pbp - pap)  # 2D triangle edge vector lengths
        lbp = vec.naive_lengths(pcp - pbp)
        lcp = vec.naive_lengths(pap - pcp)
        s = 0.5 * (lap + lbp + lcp)  # 2D triangle semi perimeter lengths
        area[..., xyz] = np.sqrt(s * (s - lap) * (s - lbp) * (s - lcp))

    return area



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['resqpy',
 'resqpy.derived_model',
 'resqpy.fault',
 'resqpy.grid',
 'resqpy.lines',
 'resqpy.model',
 'resqpy.multiprocessing',
 'resqpy.multiprocessing.wrappers',
 'resqpy.olio',
 'resqpy.olio.data',
 'resqpy.organize',
 'resqpy.property',
 'resqpy.rq_import',
 'resqpy.strata',
 'resqpy.surface',
 'resqpy.time_series',
 'resqpy.unstructured',
 'resqpy.well']

package_data = \
{'': ['*']}

install_requires = \
['h5py>=3.7,<4.0',
 'joblib>=1.2,<2.0',
 'lasio>=0.30,<0.31',
 'lxml>=4.9,<5.0',
 'numba>=0.56,<1.0',
 'numpy>=1.23,<2.0',
 'pandas>=1.5,<2.0',
 'scipy>=1.9,<2.0']

setup_kwargs = {
    'name': 'resqpy',
    'version': '3.10.1',
    'description': 'Python API for working with RESQML models',
    'long_description': '# resqpy: Python API for working with RESQML models\n\n[![License](https://img.shields.io/pypi/l/resqpy)](https://github.com/bp/resqpy/blob/master/LICENSE)\n[![Documentation Status](https://readthedocs.org/projects/resqpy/badge/?version=latest)](https://resqpy.readthedocs.io/en/latest/?badge=latest)\n[![Python CI](https://github.com/bp/resqpy/actions/workflows/ci-tests.yml/badge.svg)](https://github.com/bp/resqpy/actions/workflows/ci-tests.yml)\n![Python version](https://img.shields.io/pypi/pyversions/resqpy)\n[![PyPI](https://img.shields.io/pypi/v/resqpy)](https://badge.fury.io/py/resqpy)\n![Status](https://img.shields.io/pypi/status/resqpy)\n[![codecov](https://codecov.io/gh/bp/resqpy/branch/master/graph/badge.svg)](https://codecov.io/gh/bp/resqpy)\n\n## Introduction\n\n**resqpy** is a pure Python package which provides a programming interface (API) for\nreading, writing, and modifying reservoir models in the RESQML format. It gives\nyou the ability to work with reservoir models programmatically, without having\nto know the details of the RESQML standard.\n\nThe package is written and maintained by bp, and is made available under the MIT\nlicense as a contribution to the open-source community.\n\n**resqpy** was created by Andy Beer. For enquires about resqpy, please contact\nNathan Lane (Nathan.Lane@bp.com)\n\n### Documentation\n\nSee the complete package documentation on\n[readthedocs](https://resqpy.readthedocs.io/).\n\n### About RESQML\n\nRESQML™ is an industry initiative to provide open, non-proprietary data exchange\nstandards for reservoir characterization, earth and reservoir models. It is\ngoverned by the [Energistics\nconsortium](https://www.energistics.org/portfolio/resqml-data-standards/).\n\nResqpy provides specialized classes for a subset of the RESQML high level object\nclasses, as described in the docs. Furthermore, not all variations of these\nobject types are supported; for example, radial IJK grids are not yet catered\nfor, although the RESQML standard does allow for such grids.\n\nIt is envisaged that the code base will be expanded to include other classes of\nobject and more fully cover the options permitted by the RESQML standard.\n\nModification functionality at the moment focuses on changes to grid geometry.\n\n## Installation\n\nResqpy can be installed with pip:\n\n```bash\npip install resqpy\n```\n\nAlternatively, to install your working copy of the code in "editable" mode:\n\n```bash\npip install -e /path/to/repo/\n```\n\n## Contributing\n\nContributions of all forms are welcome and encouraged! Please feel free to open\nissues on the GitHub issue tracker, or submit Pull Requests. Instructions with how to set up your own development environment can be found at [Development environment setup](https://github.com/bp/resqpy/blob/master/docs/CONTRIBUTING.rst#development-environment-setup). Please read the\n[Contributing Guide](docs/CONTRIBUTING.rst) before submitting patches.\n',
    'author': 'BP',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/bp/resqpy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8.1,<3.11',
}


setup(**setup_kwargs)



"""Polyline and PolylineSet classes and associated functions."""

__all__ = ['Polyline', 'PolylineSet', 'load_hdf5_array', 'shift_polyline', 'flatten_polyline', 'tangents', 'spline']

from ._common import load_hdf5_array, shift_polyline, flatten_polyline, tangents, spline
from ._polyline import Polyline
from ._polyline_set import PolylineSet

# Set "module" attribute of all public objects to this path. Fixes issue #310
for _name in __all__:
    _obj = eval(_name)
    if hasattr(_obj, "__module__"):
        _obj.__module__ = __name__



"""Creating a derived resqml model from an existing one; mostly grid manipulations."""

__all__ = [
    'add_edges_per_column_property_array', 'add_faults', 'add_one_blocked_well_property', 'add_one_grid_property_array',
    'add_single_cell_grid', 'add_wells_from_ascii_file', 'add_zone_by_layer_property', 'coarsened_grid', 'copy_grid',
    'drape_to_surface', 'extract_box_for_well', 'extract_box', 'fault_throw_scaling', 'global_fault_throw_scaling',
    'gather_ensemble', 'interpolated_grid', 'local_depth_adjustment', 'refined_grid', 'tilted_grid', 'unsplit_grid',
    'zonal_grid', 'single_layer_grid', 'zone_layer_ranges_from_array'
]

from ._add_edges_per_column_property_array import add_edges_per_column_property_array
from ._add_faults import add_faults
from ._add_one_blocked_well_property import add_one_blocked_well_property
from ._add_one_grid_property_array import add_one_grid_property_array
from ._add_single_cell_grid import add_single_cell_grid
from ._add_wells_from_ascii_file import add_wells_from_ascii_file
from ._add_zone_by_layer_property import add_zone_by_layer_property
from ._coarsened_grid import coarsened_grid
from ._copy_grid import copy_grid
from ._drape_to_surface import drape_to_surface
from ._extract_box_for_well import extract_box_for_well
from ._extract_box import extract_box
from ._fault_throw_scaling import fault_throw_scaling, global_fault_throw_scaling
from ._gather_ensemble import gather_ensemble
from ._interpolated_grid import interpolated_grid
from ._local_depth_adjustment import local_depth_adjustment
from ._refined_grid import refined_grid
from ._tilted_grid import tilted_grid
from ._unsplit_grid import unsplit_grid
from ._zonal_grid import zonal_grid, single_layer_grid
from ._zone_layer_ranges_from_array import zone_layer_ranges_from_array

# Set "module" attribute of all public objects to this path.
for _name in __all__:
    _obj = eval(_name)
    if hasattr(_obj, "__module__"):
        _obj.__module__ = __name__



"""Script to construct JSON database of UOMs and property kinds.

Expects two files to exist in the same folder:
  - Energistics_Unit_of_Measure_Dictionary_V1.0.xml    (from units of measure standard)
  - Properties.xsd                                     (from resqml standard)

Creates properties.json containing parsed uoms and property kinds.
"""

import json
import math as maths
from pathlib import Path

from lxml import etree


def main():
    """Main function for parsing schema definition files and generating json file."""

    uom_dict_xml_path = Path(__file__).parent / 'Energistics_Unit_of_Measure_Dictionary_V1.0.xml'
    properties_xsd_path = Path(__file__).parent / 'Properties.xsd'
    json_path = Path(__file__).parent / 'properties.json'

    # Load xml
    assert uom_dict_xml_path.exists()
    assert properties_xsd_path.exists()
    uom_dict_root = etree.parse(str(uom_dict_xml_path)).getroot()
    properties_xsd_root = etree.parse(str(properties_xsd_path)).getroot()

    # Parse into memory
    data = dict(dimensions = parse_dimensions(uom_dict_root),
                quantities = parse_quantities(uom_dict_root),
                units = parse_units(uom_dict_root),
                prefixes = parse_prefixes(uom_dict_root),
                property_kinds = parse_property_kinds(properties_xsd_root))

    # Save to JSON
    with open(json_path, 'w', encoding = 'utf-8') as f:
        json.dump(data, f, ensure_ascii = False, indent = 2)


def parse_dimensions(root):
    """Returns a dictionary of dimensionalities."""

    dims = {}
    for node in root.find("{*}unitDimensionSet"):

        dimension = node.find('{*}dimension').text
        dims[dimension] = dict(
            name = node.find('{*}name').text,
            baseForConversion = node.find('{*}baseForConversion').text,
            canonicalUnit = node.find('{*}canonicalUnit').text,
        )
    return dims


def parse_quantities(root):
    """Returns a dictionary of quanitity classes."""

    quantities = {}
    for node in root.find("{*}quantityClassSet"):
        name = node.find('{*}name').text
        quantities[name] = dict(
            dimension = node.find('{*}dimension').text,
            baseForConversion = node.find('{*}baseForConversion').text,
        )
        alt = node.find('{*}alternativeBase')
        if alt is not None:
            quantities[name]['alternativeBase'] = alt.text

        members = []
        for member in node.findall('{*}memberUnit'):
            members.append(member.text)
        if members:
            quantities[name]["members"] = members

    return quantities


def parse_units(root):
    """Returns a dictionary of units of measure."""

    units = {}
    for node in root.find("{*}unitSet"):
        symbol = node.find('{*}symbol').text
        unit_dict = {}
        for key, typ in [
            ('name', str),
            ('dimension', str),
            ('isSI', as_bool),
            ('category', str),
            ('baseUnit', str),
            ('conversionRef', str),
            ('isExact', as_bool),
            ('A', as_numeric),
            ('B', as_numeric),
            ('C', as_numeric),
            ('D', as_numeric),
            ('conversionRef', str),
            ('description', str),
        ]:
            prop = node.find("{*}" + key)
            if prop is not None:
                unit_dict[key] = typ(prop.text)

        units[symbol] = unit_dict

    return units


def parse_prefixes(root):
    """Returns a dictionary of unit of measure prefixes."""

    prefixes = {}
    for node in root.find("{*}prefixSet"):
        name = node.find('{*}name').text
        prefixes[name] = dict(
            symbol = node.find('{*}symbol').text,
            multiplier = node.find('{*}multiplier').text,
        )
        common_name = node.find('{*}commonName')
        if common_name is not None:
            prefixes[name]['common_name'] = common_name.text

    return prefixes


def parse_property_kinds(root):
    """Return dict of valid RESQML property kinds.

    Dict keys are the valid property kinds, e.g. 'angle per time' Dict values are the description, which may be None
    """

    kind_list = get_nodes_with_name(root, 'ResqmlPropertyKind')[1]
    kind_dict = {}
    for child in kind_list:
        kind = child.get('value')
        try:
            doc = child[0][0].text
        except IndexError:
            doc = None
        kind_dict[kind] = doc
    return kind_dict


def get_nodes_with_name(root, name):
    """Find xml child node with given name."""
    for child in root:
        if child.get('name') == name:
            return child
    raise ValueError(f'{name} not found')


def as_bool(a_string):
    """Returns python bool value from string."""

    return a_string.lower() == "true"


def as_numeric(a_string):
    """Returns numeric (int or float) value from a string."""

    if a_string == 'PI':
        return maths.pi
    elif a_string == "2*PI":
        return 2 * maths.pi
    elif a_string == "4*PI":
        return 4 * maths.pi
    elif a_string.isdigit():
        return int(a_string)
    else:
        return float(a_string)


if __name__ == "__main__":
    main()



"""Submodule containing the functions relating to writing grid information."""

import logging

log = logging.getLogger(__name__)

import numpy as np

import resqpy.olio.write_hdf5 as rwh5
import resqpy.property as rprop

always_write_pillar_geometry_is_defined_array = False
always_write_cell_geometry_is_defined_array = False


def _write_hdf5_from_caches(grid,
                            file = None,
                            mode = 'a',
                            geometry = True,
                            imported_properties = None,
                            write_active = None,
                            stratigraphy = True,
                            expand_const_arrays = False):
    """Create or append to an hdf5 file.

    Writes datasets for the grid geometry (and parent grid mapping) and properties from cached arrays.
    """
    # NB: when writing a new geometry, all arrays must be set up and exist as the appropriate attributes prior to calling this function
    # if saving properties, active cell array should be added to imported_properties based on logical negation of inactive attribute
    # xml is not created here for property objects

    if write_active is None:
        write_active = geometry

    if geometry:
        grid.cache_all_geometry_arrays()

    if not file:
        file = grid.model.h5_file_name()
    h5_reg = rwh5.H5Register(grid.model)

    if stratigraphy and grid.stratigraphic_units is not None:
        h5_reg.register_dataset(grid.uuid, 'unitIndices', grid.stratigraphic_units, dtype = 'uint32')

    if geometry:
        __write_geometry(grid, h5_reg)

    if write_active and grid.inactive is not None:
        if imported_properties is None:
            imported_properties = rprop.PropertyCollection()
            imported_properties.set_support(support = grid)
        else:
            filtered_list = []
            for entry in imported_properties.imported_list:
                if (entry[2].upper() == 'ACTIVE' or entry[10] == 'active') and entry[14] == 'cells':
                    continue  # keyword or property kind
                filtered_list.append(entry)
            imported_properties.imported_list = filtered_list  # might have unintended side effects elsewhere
        active_mask = np.logical_not(grid.inactive)
        imported_properties.add_cached_array_to_imported_list(active_mask,
                                                              'active cell mask',
                                                              'ACTIVE',
                                                              discrete = True,
                                                              property_kind = 'active')

    if imported_properties is not None and imported_properties.imported_list is not None:
        for entry in imported_properties.imported_list:
            tail = 'points_patch0' if entry[18] else 'values_patch0'
            # expand constant arrays if required
            if not hasattr(imported_properties, entry[3]) and expand_const_arrays and entry[17] is not None:
                value = float(entry[17]) if isinstance(entry[17], str) else entry[17]
                assert entry[14] == 'cells' and entry[15] == 1
                imported_properties.__dict__[entry[3]] = np.full(grid.extent_kji, value)
            if hasattr(imported_properties, entry[3]):  # otherwise constant array not being expanded
                h5_reg.register_dataset(entry[0], tail, imported_properties.__dict__[entry[3]])
            if entry[10] == 'active':
                grid.active_property_uuid = entry[0]
    h5_reg.write(file, mode = mode)


def __write_geometry(grid, h5_reg):
    if always_write_pillar_geometry_is_defined_array or not grid.geometry_defined_for_all_pillars(cache_array = True):
        if not hasattr(grid, 'array_pillar_geometry_is_defined') or grid.array_pillar_geometry_is_defined is None:
            grid.array_pillar_geometry_is_defined = np.full((grid.nj + 1, grid.ni + 1), True, dtype = bool)
        h5_reg.register_dataset(grid.uuid,
                                'PillarGeometryIsDefined',
                                grid.array_pillar_geometry_is_defined,
                                dtype = 'uint8')
    if always_write_cell_geometry_is_defined_array or not grid.geometry_defined_for_all_cells(cache_array = True):
        if not hasattr(grid, 'array_cell_geometry_is_defined') or grid.array_cell_geometry_is_defined is None:
            grid.array_cell_geometry_is_defined = np.full((grid.nk, grid.nj, grid.ni), True, dtype = bool)
        h5_reg.register_dataset(grid.uuid,
                                'CellGeometryIsDefined',
                                grid.array_cell_geometry_is_defined,
                                dtype = 'uint8')
    # todo: PillarGeometryIsDefined ?
    h5_reg.register_dataset(grid.uuid, 'Points', grid.points_cached)
    if grid.has_split_coordinate_lines:
        h5_reg.register_dataset(grid.uuid, 'PillarIndices', grid.split_pillar_indices_cached, dtype = 'uint32')
        h5_reg.register_dataset(grid.uuid,
                                'ColumnsPerSplitCoordinateLine/elements',
                                grid.cols_for_split_pillars,
                                dtype = 'uint32')
        h5_reg.register_dataset(grid.uuid,
                                'ColumnsPerSplitCoordinateLine/cumulativeLength',
                                grid.cols_for_split_pillars_cl,
                                dtype = 'uint32')
    if grid.k_gaps:
        assert grid.k_gap_after_array is not None
        h5_reg.register_dataset(grid.uuid, 'GapAfterLayer', grid.k_gap_after_array, dtype = 'uint8')
    if grid.parent_window is not None:
        for axis in range(3):
            if grid.parent_window.fine_extent_kji[axis] == grid.parent_window.coarse_extent_kji[axis]:
                continue  # one-to-noe mapping
            # reconstruct hdf5 arrays from FineCoarse object and register for write
            if grid.parent_window.constant_ratios[axis] is not None:
                if grid.is_refinement:
                    pcpi = np.array([grid.parent_window.coarse_extent_kji[axis]], dtype = int)  # ParentCountPerInterval
                    ccpi = np.array([grid.parent_window.fine_extent_kji[axis]], dtype = int)  # ChildCountPerInterval
                else:
                    pcpi = np.array([grid.parent_window.fine_extent_kji[axis]], dtype = int)
                    ccpi = np.array([grid.parent_window.coarse_extent_kji[axis]], dtype = int)
            else:
                if grid.is_refinement:
                    interval_count = grid.parent_window.coarse_extent_kji[axis]
                    pcpi = np.ones(interval_count, dtype = int)
                    ccpi = np.array(grid.parent_window.vector_ratios[axis], dtype = int)
                else:
                    interval_count = grid.parent_window.fine_extent_kji[axis]
                    pcpi = np.array(grid.parent_window.vector_ratios[axis], dtype = int)
                    ccpi = np.ones(interval_count, dtype = int)
            h5_reg.register_dataset(grid.uuid, 'KJI'[axis] + 'Regrid/ParentCountPerInterval', pcpi)
            h5_reg.register_dataset(grid.uuid, 'KJI'[axis] + 'Regrid/ChildCountPerInterval', ccpi)
            if grid.is_refinement and not grid.parent_window.equal_proportions[axis]:
                child_cell_weights = np.concatenate(grid.parent_window.vector_proportions[axis])
                h5_reg.register_dataset(grid.uuid, 'KJI'[axis] + 'Regrid/ChildCellWeights', child_cell_weights)
